# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

# Adapted from https://github.com/Lightning-AI/metrics/blob/1c42f6643f9241089e55a4d899f14da480021e16/torchmetrics/detection/map.py
# Current versions of MAP in torchmetrics are incorrect or slow as of 9/21/22.
# Relevant issues:
# https://github.com/Lightning-AI/metrics/issues/1024
# https://github.com/Lightning-AI/metrics/issues/1164

"""MAP torchmetric for object detection."""
#type: ignore
import logging
import sys
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Sequence, Union

import torch
from torch import Tensor
from torchmetrics.metric import Metric
from torchvision.ops import box_convert

from composer.utils import MissingConditionalImportError

__all__ = ['MAP']

log = logging.getLogger(__name__)


@dataclass
class MAPMetricResults:
    """Dataclass to wrap the final mAP results."""
    map: Tensor
    map_50: Tensor
    map_75: Tensor
    map_small: Tensor
    map_medium: Tensor
    map_large: Tensor
    mar_1: Tensor
    mar_10: Tensor
    mar_100: Tensor
    mar_small: Tensor
    mar_medium: Tensor
    mar_large: Tensor
    map_per_class: Tensor
    mar_100_per_class: Tensor

    def __getitem__(self, key: str) -> Union[Tensor, List[Tensor]]:
        return getattr(self, key)


# noinspection PyMethodMayBeStatic
class WriteToLog:
    """Logging class to move logs to log.debug()."""

    def write(self, buf: str) -> None:  # skipcq: PY-D0003, PYL-R0201
        for line in buf.rstrip().splitlines():
            log.debug(line.rstrip())

    def flush(self) -> None:  # skipcq: PY-D0003, PYL-R0201
        for handler in log.handlers:
            handler.flush()

    def close(self) -> None:  # skipcq: PY-D0003, PYL-R0201
        for handler in log.handlers:
            handler.close()


class _hide_prints:
    """Internal helper context to suppress the default output of the pycocotools package."""

    def __init__(self) -> None:
        self._original_stdout = None

    def __enter__(self) -> None:
        self._original_stdout = sys.stdout  # type: ignore
        sys.stdout = WriteToLog()  # type: ignore

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:  # type: ignore
        sys.stdout.close()
        sys.stdout = self._original_stdout  # type: ignore


def _input_validator(preds: List[Dict[str, torch.Tensor]], targets: List[Dict[str, torch.Tensor]]) -> None:
    """Ensure the correct input format of `preds` and `targets`."""
    if not isinstance(preds, Sequence):
        raise ValueError('Expected argument `preds` to be of type List')
    if not isinstance(targets, Sequence):
        raise ValueError('Expected argument `target` to be of type List')
    if len(preds) != len(targets):
        raise ValueError('Expected argument `preds` and `target` to have the same length')

    for k in ['boxes', 'scores', 'labels']:
        if any(k not in p for p in preds):
            raise ValueError(f'Expected all dicts in `preds` to contain the `{k}` key')

    for k in ['boxes', 'labels']:
        if any(k not in p for p in targets):
            raise ValueError(f'Expected all dicts in `target` to contain the `{k}` key')

    if any(type(pred['boxes']) is not torch.Tensor for pred in preds):
        raise ValueError('Expected all boxes in `preds` to be of type torch.Tensor')
    if any(type(pred['scores']) is not torch.Tensor for pred in preds):
        raise ValueError('Expected all scores in `preds` to be of type torch.Tensor')
    if any(type(pred['labels']) is not torch.Tensor for pred in preds):
        raise ValueError('Expected all labels in `preds` to be of type torch.Tensor')
    if any(type(target['boxes']) is not torch.Tensor for target in targets):
        raise ValueError('Expected all boxes in `target` to be of type torch.Tensor')
    if any(type(target['labels']) is not torch.Tensor for target in targets):
        raise ValueError('Expected all labels in `target` to be of type torch.Tensor')

    for i, item in enumerate(targets):
        if item['boxes'].size(0) != item['labels'].size(0):
            raise ValueError(
                f'Input boxes and labels of sample {i} in targets have a'
                f" different length (expected {item['boxes'].size(0)} labels, got {item['labels'].size(0)})")
    for i, item in enumerate(preds):
        if item['boxes'].size(0) != item['labels'].size(0) != item['scores'].size(0):
            raise ValueError(f'Input boxes, labels and scores of sample {i} in preds have a'
                             f" different length (expected {item['boxes'].size(0)} labels and scores,"
                             f" got {item['labels'].size(0)} labels and {item['scores'].size(0)})")


class MAP(Metric):
    """Computes the Mean-Average-Precision (mAP) and Mean-Average-Recall (mAR) for object detection predictions.

    Optionally, the mAP and mAR values can be calculated per class.
    Predicted boxes and targets have to be in Pascal VOC format \
    (xmin-top left, ymin-top left, xmax-bottom right, ymax-bottom right).
    See the :meth:`update` method for more information about the input format to this metric.
    See `this blog <https://jonathan-hui.medium.com/map-mean-average-precision-for-object-detection-45c121a31173>`_ for more details on (mAP)
    and (mAR).

    .. warning:: This metric is a wrapper for the `pycocotools <https://github.com/cocodataset/cocoapi/tree/master/PythonAPI/pycocotools>`_,
        which is a standard implementation for the mAP metric for object detection. Using this metric
        therefore requires you to have `pycocotools` installed. Please install with ``pip install pycocotools``

    .. warning:: As the pycocotools library cannot deal with tensors directly, all results have to be transfered
        to the CPU, this may have an performance impact on your training.

    Args:
        class_metrics (bool, optional): Option to enable per-class metrics for mAP and mAR_100. Has a performance impact. Default: ``False``.
        compute_on_step (bool, optional): Forward only calls ``update()`` and return ``None`` if this is set to ``False``. Default: ``False``.
        dist_sync_on_step (bool, optional): Synchronize metric state across processes at each ``forward()`` before returning the value at the step. Default: ``False``.
        process_group (any, optional): Specify the process group on which synchronization is called. Default: ``None`` (which selects the entire world).
        dist_sync_fn (callable, optional): Callback that performs the allgather operation on the metric state. When ``None``, DDP will be used to perform the all_gather. Default: ``None``.

    Raises:
        ValueError: If ``class_metrics`` is not a boolean.
    """

    # Have default behavior for this complex metric
    full_state_update = True

    def __init__(
            self,
            class_metrics: bool = False,
            compute_on_step: bool = True,
            dist_sync_on_step: bool = False,
            process_group: Optional[Any] = None,
            dist_sync_fn: Callable = None,  # type: ignore
    ) -> None:  # type: ignore
        super().__init__(
            compute_on_step=compute_on_step,
            dist_sync_on_step=dist_sync_on_step,
            process_group=process_group,
            dist_sync_fn=dist_sync_fn,
        )
        try:
            from pycocotools.coco import COCO
            from pycocotools.cocoeval import COCOeval
        except ImportError as e:
            raise MissingConditionalImportError(extra_deps_group='coco', conda_package='pycocotools') from e

        self.COCO = COCO
        self.COCOeval = COCOeval

        if not isinstance(class_metrics, bool):
            raise ValueError('Expected argument `class_metrics` to be a boolean')
        self.class_metrics = class_metrics

        self.add_state('detection_boxes', default=[], dist_reduce_fx=None)
        self.add_state('detection_scores', default=[], dist_reduce_fx=None)
        self.add_state('detection_labels', default=[], dist_reduce_fx=None)
        self.add_state('groundtruth_boxes', default=[], dist_reduce_fx=None)
        self.add_state('groundtruth_labels', default=[], dist_reduce_fx=None)

    def update(self, preds: List[Dict[str, Tensor]], target: List[Dict[str, Tensor]]) -> None:  # type: ignore
        """Add detections and groundtruth to the metric.

        Args:
            preds (list[Dict[str, ~torch.Tensor]]): A list of dictionaries containing the key-values:

                    ``boxes`` (torch.FloatTensor): [num_boxes, 4] predicted boxes of the format [xmin, ymin, xmax, ymax] in absolute image coordinates.

                    ``scores`` (torch.FloatTensor): of shape [num_boxes] containing detection scores for the boxes.

                    ``labels`` (torch.IntTensor): of shape [num_boxes] containing 0-indexed detection classes for the boxes.

            target (list[Dict[str, ~torch.Tensor]]): A list of dictionaries containing the key-values:

                    ``boxes`` (torch.FloatTensor): [num_boxes, 4] ground truth boxes of the format [xmin, ymin, xmax, ymax] in absolute image coordinates.

                    ``labels`` (torch.IntTensor): of shape [num_boxes] containing 1-indexed groundtruth classes for the boxes.

        Raises:
            ValueError: If ``preds`` and ``target`` are not of the same length.
            ValueError: If any of ``preds.boxes``, ``preds.scores`` and ``preds.labels`` are not of the same length.
            ValueError: If any of ``target.boxes`` and ``target.labels`` are not of the same length.
            ValueError: If any box is not type float and of length 4.
            ValueError: If any class is not type int and of length 1.
            ValueError: If any score is not type float and of length 1.
        """
        _input_validator(preds, target)

        for item in preds:
            self.detection_boxes.append(item['boxes'])  # type: ignore
            self.detection_scores.append(item['scores'])  # type: ignore
            self.detection_labels.append(item['labels'])  # type: ignore

        for item in target:
            self.groundtruth_boxes.append(item['boxes'])  # type: ignore
            self.groundtruth_labels.append(item['labels'])  # type: ignore

    def compute(self) -> dict:
        """Compute the Mean-Average-Precision (mAP) and Mean-Average-Recall (mAR) scores.

        All detections added in the ``update()`` method are included.

        Note:
            Main `map` score is calculated with @[ IoU=0.50:0.95 | area=all | maxDets=100 ]

        Returns:
            MAPMetricResults (dict): containing:

                ``map`` (torch.Tensor): map at 95 iou.

                ``map_50`` (torch.Tensor): map at 50 iou.

                ``map_75`` (torch.Tensor): map at 75 iou.

                ``map_small`` (torch.Tensor): map at 95 iou for small objects.

                ``map_medium`` (torch.Tensor): map at 95 iou for medium objects.

                ``map_large`` (torch.Tensor): map at 95 iou for large objects.

                ``mar_1`` (torch.Tensor): mar at 1 max detection.

                ``mar_10`` (torch.Tensor): mar at 10 max detections.

                ``mar_100`` (torch.Tensor): mar at 100 max detections.

                ``mar_small`` (torch.Tensor): mar at 100 max detections for small objects.

                ``mar_medium`` (torch.Tensor): mar at 100 max detections for medium objects.

                ``mar_large`` (torch.Tensor): mar at 100 max detections for large objects.

                ``map_per_class`` (torch.Tensor) (-1 if class metrics are disabled): map value for each class.

                ``mar_100_per_class`` (torch.Tensor) (-1 if class metrics are disabled): mar at 100 detections for each class.
        """
        coco_target, coco_preds = self.COCO(), self.COCO()  # type: ignore
        coco_target.dataset = self._get_coco_format(self.groundtruth_boxes, self.groundtruth_labels)  # type: ignore
        coco_preds.dataset = self._get_coco_format(  # type: ignore
            self.detection_boxes,  # type: ignore
            self.detection_labels,  # type: ignore
            self.detection_scores)  # type: ignore

        with _hide_prints():
            coco_target.createIndex()
            coco_preds.createIndex()
            coco_eval = self.COCOeval(coco_target, coco_preds, 'bbox')  # type: ignore
            coco_eval.evaluate()
            coco_eval.accumulate()
            coco_eval.summarize()
            stats = coco_eval.stats

        map_per_class_values: Tensor = torch.Tensor([-1])
        mar_100_per_class_values: Tensor = torch.Tensor([-1])
        # if class mode is enabled, evaluate metrics per class
        if self.class_metrics:
            map_per_class_list = []
            mar_100_per_class_list = []
            for class_id in torch.cat(self.detection_labels +
                                      self.groundtruth_labels).unique().cpu().tolist():  # type: ignore
                coco_eval.params.catIds = [class_id]
                with _hide_prints():
                    coco_eval.evaluate()
                    coco_eval.accumulate()
                    coco_eval.summarize()
                    class_stats = coco_eval.stats

                map_per_class_list.append(torch.Tensor([class_stats[0]]))
                mar_100_per_class_list.append(torch.Tensor([class_stats[8]]))
            map_per_class_values = torch.Tensor(map_per_class_list)
            mar_100_per_class_values = torch.Tensor(mar_100_per_class_list)

        metrics = MAPMetricResults(
            map=torch.Tensor([stats[0]]),
            map_50=torch.Tensor([stats[1]]),
            map_75=torch.Tensor([stats[2]]),
            map_small=torch.Tensor([stats[3]]),
            map_medium=torch.Tensor([stats[4]]),
            map_large=torch.Tensor([stats[5]]),
            mar_1=torch.Tensor([stats[6]]),
            mar_10=torch.Tensor([stats[7]]),
            mar_100=torch.Tensor([stats[8]]),
            mar_small=torch.Tensor([stats[9]]),
            mar_medium=torch.Tensor([stats[10]]),
            mar_large=torch.Tensor([stats[11]]),
            map_per_class=map_per_class_values,
            mar_100_per_class=mar_100_per_class_values,
        )
        return metrics.__dict__

    def _get_coco_format(self,
                         boxes: List[torch.Tensor],
                         labels: List[torch.Tensor],
                         scores: Optional[List[torch.Tensor]] = None) -> Dict:
        """Transforms and returns all cached targets or predictions in COCO format.

        Format is defined at https://cocodataset.org/#format-data.
        """
        images = []
        annotations = []
        annotation_id = 1  # has to start with 1, otherwise COCOEval results are wrong

        boxes = [box_convert(box, in_fmt='xyxy', out_fmt='xywh') for box in boxes]  # type: ignore
        for image_id, (image_boxes, image_labels) in enumerate(zip(boxes, labels)):
            image_boxes = image_boxes.cpu().tolist()
            image_labels = image_labels.cpu().tolist()

            images.append({'id': image_id})
            for k, (image_box, image_label) in enumerate(zip(image_boxes, image_labels)):
                if len(image_box) != 4:
                    raise ValueError(
                        f'Invalid input box of sample {image_id}, element {k} (expected 4 values, got {len(image_box)})'
                    )

                if type(image_label) != int:
                    raise ValueError(f'Invalid input class of sample {image_id}, element {k}'
                                     f' (expected value of type integer, got type {type(image_label)})')

                annotation = {
                    'id': annotation_id,
                    'image_id': image_id,
                    'bbox': image_box,
                    'category_id': image_label,
                    'area': image_box[2] * image_box[3],
                    'iscrowd': 0,
                }
                if scores is not None:  # type: ignore
                    score = scores[image_id][k].cpu().tolist()  # type: ignore
                    if type(score) != float:  # type: ignore
                        raise ValueError(f'Invalid input score of sample {image_id}, element {k}'
                                         f' (expected value of type float, got type {type(score)})')
                    annotation['score'] = score
                annotations.append(annotation)
                annotation_id += 1

        classes = [{
            'id': i,
            'name': str(i)
        } for i in torch.cat(self.detection_labels + self.groundtruth_labels).unique().cpu().tolist()]  # type: ignore
        return {'images': images, 'annotations': annotations, 'categories': classes}



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Logs to a file."""

from __future__ import annotations

import os
import queue
import sys
import textwrap
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, TextIO

from composer.loggers.logger import Logger, format_log_data_value
from composer.loggers.logger_destination import LoggerDestination
from composer.utils import FORMAT_NAME_WITH_DIST_TABLE, format_name_with_dist

if TYPE_CHECKING:
    from composer.core import State

__all__ = ['FileLogger']


class FileLogger(LoggerDestination):  # noqa: D101
    __doc__ = f"""Log data to a file.

    Example usage:
        .. testcode::

            from composer.loggers import FileLogger
            from composer.trainer import Trainer
            file_logger = FileLogger(
                filename="{{run_name}}/logs-rank{{rank}}.txt",
                buffer_size=1,
                flush_interval=50
            )
            trainer = Trainer(
                ...,
                loggers=[file_logger]
            )

        .. testcleanup::

            import os

            trainer.engine.close()

            path = os.path.join(trainer.state.run_name, "logs-rank0.txt")
            try:
                os.remove(file_logger.filename)
            except FileNotFoundError as e:
                pass

    Example output::

        [FIT][step=2]: {{ "logged_metric": "logged_value", }}
        [EPOCH][step=2]: {{ "logged_metric": "logged_value", }}
        [BATCH][step=2]: {{ "logged_metric": "logged_value", }}
        [EPOCH][step=3]: {{ "logged_metric": "logged_value", }}


    Args:
        filename (str, optional): Format string for the filename.

            The following format variables are available:

            {textwrap.indent(FORMAT_NAME_WITH_DIST_TABLE, prefix='            ')}

            .. note::

                When training with multiple devices (i.e. GPUs), ensure that ``'{{rank}}'`` appears in the format.
                Otherwise, multiple processes may attempt to write to the same file.

            Consider the following example when using default value of '{{run_name}}/logs-rank{{rank}}.txt':

            >>> file_logger = FileLogger(filename='{{run_name}}/logs-rank{{rank}}.txt')
            >>> trainer = Trainer(loggers=[file_logger], run_name='my-awesome-run')
            >>> file_logger.filename
            'my-awesome-run/logs-rank0.txt'

            Default: `'{{run_name}}/logs-rank{{rank}}.txt'`

        remote_file_name (str, optional): Format string for the logfile's name.

            The logfile will be periodically logged (according to the ``flush_interval``) as a file.
            The file name will be determined by this format string.

            .. seealso:: :doc:`Uploading Files</trainer/file_uploading>` for notes for file uploading.

            The same format variables for ``filename`` are available. Setting this parameter to ``None``
            (the default) will use the same format string as ``filename``. It is sometimes helpful to deviate
            from this default. For example, when ``filename`` contains an absolute path, it is recommended to
            set this parameter explicitely, so the absolute path does not appear in any remote file stores.

            Leading slashes (``'/'``) will be stripped.

            Default: ``None`` (which uses the same format string as ``filename``)
        capture_stdout (bool, optional): Whether to include the ``stdout``in ``filename``. (default: ``True``)
        capture_stderr (bool, optional): Whether to include the ``stderr``in ``filename``. (default: ``True``)
        buffer_size (int, optional): Buffer size. See :py:func:`open`.
            Default: ``1`` for line buffering.
        log_traces (bool, optional): Whether to log algorithm traces. See :class:`~.Engine` for more detail.
        flush_interval (int, optional): How frequently to flush the log to the file in batches
            Default: ``100``.
        overwrite (bool, optional): Whether to overwrite an existing logfile. (default: ``False``)
    """

    def __init__(
        self,
        filename: str = '{run_name}/logs-rank{rank}.txt',
        remote_file_name: Optional[str] = None,
        *,
        capture_stdout: bool = True,
        capture_stderr: bool = True,
        buffer_size: int = 1,
        log_traces: bool = True,
        flush_interval: int = 100,
        overwrite: bool = False,
    ) -> None:
        self.filename_format = filename
        if remote_file_name is None:
            remote_file_name = filename.replace(os.path.sep, '/')
        self.remote_file_name_format = remote_file_name
        self.buffer_size = buffer_size
        self.should_log_traces = log_traces
        self.flush_interval = flush_interval
        self.is_batch_interval = False
        self.is_epoch_interval = False
        self.file: Optional[TextIO] = None
        self.overwrite = overwrite,
        self._queue: queue.Queue[str] = queue.Queue()
        self._run_name = None
        # Track whether the next line is on a newline
        # (and if so, then the prefix should be appended)
        self._is_newline = True
        self._closed = False

        if capture_stdout:
            sys.stdout.write = self._get_new_writer('[stdout]: ', sys.stdout.write)

        if capture_stderr:
            sys.stderr.write = self._get_new_writer('[stderr]: ', sys.stderr.write)

    def _get_new_writer(self, prefix: str, original_writer: Callable[[str], int]):
        """Returns a writer that intercepts calls to the ``original_writer``."""

        def new_write(s: str) -> int:
            if not self._closed:
                self.write(prefix, s)
            return original_writer(s)

        return new_write

    @property
    def filename(self) -> str:
        """The filename for the logfile."""
        if self._run_name is None:
            raise RuntimeError('The run name is not set. The engine should have been set on Event.INIT')
        name = format_name_with_dist(self.filename_format, run_name=self._run_name)

        return name

    @property
    def remote_file_name(self) -> str:
        """The remote file name for the logfile."""
        if self._run_name is None:
            raise RuntimeError('The run name is not set. The engine should have been set on Event.INIT')
        name = format_name_with_dist(self.remote_file_name_format, run_name=self._run_name)

        name.lstrip('/')

        return name

    def epoch_start(self, state: State, logger: Logger) -> None:
        # Flush any log calls that occurred during INIT or FIT_START
        self._flush_file(logger)

    def log_traces(self, traces: Dict[str, Any]):
        if self.should_log_traces:
            for trace_name, trace in traces.items():
                trace_str = format_log_data_value(trace)
                self.write(
                    f'[trace]: {trace_name}:',
                    trace_str + '\n',
                )

    def log_metrics(self, metrics: Dict[str, float], step: Optional[int] = None) -> None:
        for metric_name, metric in metrics.items():
            metric_str = format_log_data_value(metric)
            self.write(
                f'[metric][batch={step}]: ',
                f'{metric_name}: {metric_str} \n',
            )

    def log_hyperparameters(self, hyperparameters: Dict[str, Any]):
        for hparam_name, hparam_value in hyperparameters.items():
            hparam_str = format_log_data_value(hparam_value)
            self.write(
                f'[hyperparameter]: ',
                f'{hparam_name}: {hparam_str} \n',
            )

    def init(self, state: State, logger: Logger) -> None:
        del logger  # unused
        self._is_newline = True
        self._run_name = state.run_name
        if self.file is not None:
            raise RuntimeError('The file logger is already initialized')
        file_dirname = os.path.dirname(self.filename)
        if file_dirname:
            os.makedirs(file_dirname, exist_ok=True)
        mode = 'w+' if self.overwrite else 'x+'
        self.file = open(self.filename, mode, buffering=self.buffer_size)
        self._flush_queue()

    def batch_end(self, state: State, logger: Logger) -> None:
        assert self.file is not None
        if int(state.timestamp.batch) % self.flush_interval == 0:
            self._flush_file(logger)

    def eval_start(self, state: State, logger: Logger) -> None:
        # Flush any log calls that occurred during INIT when using the trainer in eval-only mode
        self._flush_file(logger)

    def epoch_end(self, state: State, logger: Logger) -> None:
        if int(state.timestamp.epoch) % self.flush_interval == 0:
            self._flush_file(logger)

    def write(self, prefix: str, s: str):
        """Write to the logfile.

        .. note::

            If the ``write`` occurs before the :attr:`.Event.INIT` event,
            the write will be enqueued, as the file is not yet open.

        Args:
            prefix (str): A prefix for each line in the logfile.
            s (str): The string to write. Each line will be prefixed with ``prefix``.
        """
        formatted_lines = []
        for line in s.splitlines(True):
            if self._is_newline:
                # Only print the prefix if it is a newline
                # and the line is not empty
                if line == os.linesep:
                    formatted_lines.append(line)
                else:
                    formatted_lines.append(f'{prefix}{line}')
                self._is_newline = False
            else:
                # Otherwise, append the line without the prefix
                formatted_lines.append(line)
            if line.endswith(os.linesep):
                # if the line ends with newline, record that the next
                # line should start with the prefix
                self._is_newline = True
        formatted_s = ''.join(formatted_lines)
        if self.file is None:
            self._queue.put_nowait(formatted_s)
        else:
            # Flush the queue, so all prints will be in order
            self._flush_queue()
            # Then, write to the file
            print(formatted_s, file=self.file, flush=False, end='')

    def _flush_queue(self):
        while True:
            try:
                s = self._queue.get_nowait()
            except queue.Empty:
                break
            print(s, file=self.file, flush=False, end='')

    def _flush_file(self, logger: Logger) -> None:
        assert self.file is not None

        self._flush_queue()

        self.file.flush()
        os.fsync(self.file.fileno())
        logger.upload_file(self.remote_file_name, self.file.name, overwrite=True)

    def fit_end(self, state: State, logger: Logger) -> None:
        # Flush the file on fit_end, in case if was not flushed on epoch_end and the trainer is re-used
        # (which would defer when `self.close()` would be invoked)
        self._flush_file(logger)

    def close(self, state: State, logger: Logger) -> None:
        del state  # unused
        self._closed = True  # Stop intercepting calls to stdout/stderr
        if self.file is not None:
            self._flush_file(logger)
            self.file.close()
            self.file = None



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

import pytest
import torch
import torch.distributed
from torch.utils.data import DataLoader

from composer import Trainer
from composer.core import Precision
from composer.models import composer_resnet_cifar
from tests.common import RandomImageDataset

try:
    import transformer_engine.pytorch as te
    del te
    te_installed = True
except ImportError:
    te_installed = False


def get_trainer(precision: Precision) -> Trainer:

    return Trainer(
        model=composer_resnet_cifar('resnet_9'),
        train_dataloader=DataLoader(
            dataset=RandomImageDataset(size=1024),
            batch_size=512,
            persistent_workers=False,
            num_workers=0,
        ),
        eval_dataloader=DataLoader(
            dataset=RandomImageDataset(size=1024),
            batch_size=512,
            persistent_workers=False,
            num_workers=0,
        ),
        precision=precision,
        max_duration='1ep',
        eval_interval='1ep',
        train_subset_num_batches=1,
    )


def fit_and_measure_memory(precision) -> int:
    trainer = get_trainer(precision)
    torch.cuda.empty_cache()
    torch.cuda.reset_peak_memory_stats()

    trainer.fit()

    return torch.cuda.max_memory_allocated()


def eval_and_measure_memory(precision) -> int:
    trainer = get_trainer(precision)
    torch.cuda.empty_cache()
    torch.cuda.reset_peak_memory_stats()

    trainer.eval()

    return torch.cuda.max_memory_allocated()


def predict_and_measure_memory(precision) -> int:
    trainer = get_trainer(precision)
    torch.cuda.empty_cache()
    torch.cuda.reset_peak_memory_stats()

    trainer.predict(dataloader=trainer.state.evaluators[0].dataloader)

    return torch.cuda.max_memory_allocated()


@pytest.mark.gpu
@pytest.mark.parametrize('precision', [Precision.AMP_FP16, Precision.AMP_BF16])
def test_train_precision_memory(precision: Precision):
    memory_fp32 = fit_and_measure_memory(Precision.FP32)
    memory_half = fit_and_measure_memory(precision)
    assert memory_half < 0.7 * memory_fp32


@pytest.mark.gpu
@pytest.mark.parametrize('precision', [Precision.AMP_FP16, Precision.AMP_BF16])
def test_eval_precision_memory(precision: Precision):
    memory_fp32 = eval_and_measure_memory(Precision.FP32)
    memory_half = eval_and_measure_memory(precision)
    assert memory_half < 0.95 * memory_fp32


@pytest.mark.gpu
@pytest.mark.parametrize('precision', [Precision.AMP_FP16, Precision.AMP_BF16])
def test_predict_precision_memory(precision: Precision):
    memory_fp32 = predict_and_measure_memory(Precision.FP32)
    memory_half = predict_and_measure_memory(precision)
    assert memory_half < 0.95 * memory_fp32


@pytest.mark.gpu
def test_amp_fp8_path():
    trainer = get_trainer(Precision.AMP_FP8)
    if te_installed:
        if torch.cuda.get_device_capability()[0] < 9:
            with pytest.raises(RuntimeError, match='AMP_FP8 precision is used but current device does not support it'):
                trainer.fit()
        else:
            trainer.fit()
    else:
        with pytest.raises(ImportError, match='AMP_FP8 precision is used but TransformerEngine is not installed'):
            trainer.fit()



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Helpers for working with files."""

from __future__ import annotations

import logging
import os
import pathlib
import re
import tempfile
import uuid
import warnings
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse

import requests
import tqdm

from composer.utils import dist
from composer.utils.iter_helpers import iterate_with_callback
from composer.utils.object_store import LibcloudObjectStore, ObjectStore, OCIObjectStore, S3ObjectStore

if TYPE_CHECKING:
    from composer.core import Timestamp
    from composer.loggers import LoggerDestination, RemoteUploaderDownloader

log = logging.getLogger(__name__)

__all__ = [
    'get_file', 'ensure_folder_is_empty', 'ensure_folder_has_no_conflicting_files', 'format_name_with_dist',
    'format_name_with_dist_and_time', 'is_tar', 'create_symlink_file', 'maybe_create_object_store_from_uri',
    'maybe_create_remote_uploader_downloader_from_uri', 'parse_uri'
]


def _get_dist_config(strict: bool = True) -> Dict[str, Any]:
    """Returns a dict of distributed settings (rank, world_size, etc.).

    If ``strict=True``, will error if a setting is not available (e.g. the
    environment variable is not set). Otherwise, will only return settings
    that are available.
    """
    settings = {
        'rank': dist.get_global_rank,
        'local_rank': dist.get_local_rank,
        'world_size': dist.get_world_size,
        'local_world_size': dist.get_local_world_size,
        'node_rank': dist.get_node_rank,
    }

    dist_config = {}
    for name, func in settings.items():
        try:
            value = func()
        except dist.MissingEnvironmentError as e:
            if strict:
                raise e
        else:
            dist_config[name] = value

    return dist_config


def is_tar(name: Union[str, pathlib.Path]) -> bool:
    """Returns whether ``name`` has a tar-like extension.

    Args:
        name (str | pathlib.Path): The name to check.

    Returns:
        bool: Whether ``name`` is a tarball.
    """
    return any(str(name).endswith(x) for x in ('.tar', '.tgz', '.tar.gz', '.tar.bz2', '.tar.lzma'))


def ensure_folder_is_empty(folder_name: Union[str, pathlib.Path]):
    """Ensure that the given folder is empty.

    Hidden files and folders (those beginning with ``.``) and ignored. Sub-folders are checked recursively.

    Args:
        folder_name (str | pathlib.Path): The folder to ensure is empty.

    Raises:
        FileExistsError: If ``folder_name`` contains any non-hidden files, recursively.
    """
    for root, dirs, files in os.walk(folder_name, topdown=True):
        # Filter out hidden folders
        dirs[:] = (x for x in dirs if not x.startswith('.'))
        for file in files:
            if not file.startswith('.'):
                raise FileExistsError(f'{folder_name} is not empty; {os.path.join(root, file)} exists.')


def ensure_folder_has_no_conflicting_files(folder_name: Union[str, pathlib.Path], filename: str, timestamp: Timestamp):
    """Ensure that the given folder does not have any files conflicting with the ``filename`` format string.

    If any filename is formatted with a timestamp where the epoch, batch, sample, or token counts are after
    ``timestamp``, a ``FileExistsError`` will be raised.
    If ``filename`` and occurs later than ``timestamp``, raise a ``FileExistsError``.

    Args:
        folder_name (str | pathlib.Path): The folder to inspect.
        filename (str): The pattern string for potential files.
        timestamp (Timestamp): Ignore any files that occur before the provided timestamp.

    Raises:
        FileExistsError: If ``folder_name`` contains any files matching the ``filename`` template before ``timestamp``.
    """
    # Prepare regex pattern by replacing f-string formatting with regex.
    pattern = f'^{filename}$'

    # Format time vars for regex match
    for unit in ['epoch', 'batch', 'sample', 'token', 'batch_in_epoch', 'sample_in_epoch', 'token_in_epoch']:
        if unit in filename:
            pattern = pattern.replace(f'{{{unit}}}', f'(?P<{unit}>\\d+)')

    # Format rank information
    pattern = pattern.format(**_get_dist_config(strict=False))

    template = re.compile(pattern)

    for file in os.listdir(folder_name):
        match = template.match(file)

        if match is not None:
            match = match.groupdict()
            for unit, value in match.items():
                if unit.endswith('_in_epoch'):
                    if 'epoch' not in match:
                        raise ValueError(f'{filename} has {{unit}} but not {{epoch}}. Add {{epoch}} for uniqueness.')
                    if int(match['epoch']) != timestamp.epoch:
                        continue  # only check _in_epoch if both files have same epoch count

                if int(value) > int(getattr(timestamp, unit)):
                    raise FileExistsError(
                        f'{os.path.join(folder_name, file)} may conflict with a future checkpoint of the current run.'
                        'Please delete that file, change to a new folder, or set overwrite=True.')


FORMAT_NAME_WITH_DIST_TABLE = """
+------------------------+-------------------------------------------------------+
| Variable               | Description                                           |
+========================+=======================================================+
| ``{run_name}``         | The name of the training run. See                     |
|                        | :attr:`.Logger.run_name`.                             |
+------------------------+-------------------------------------------------------+
| ``{rank}``             | The global rank, as returned by                       |
|                        | :func:`~composer.utils.dist.get_global_rank`.         |
+------------------------+-------------------------------------------------------+
| ``{local_rank}``       | The local rank of the process, as returned by         |
|                        | :func:`~composer.utils.dist.get_local_rank`.          |
+------------------------+-------------------------------------------------------+
| ``{world_size}``       | The world size, as returned by                        |
|                        | :func:`~composer.utils.dist.get_world_size`.          |
+------------------------+-------------------------------------------------------+
| ``{local_world_size}`` | The local world size, as returned by                  |
|                        | :func:`~composer.utils.dist.get_local_world_size`.    |
+------------------------+-------------------------------------------------------+
| ``{node_rank}``        | The node rank, as returned by                         |
|                        | :func:`~composer.utils.dist.get_node_rank`.           |
+------------------------+-------------------------------------------------------+
"""


def format_name_with_dist(format_str: str, run_name: str, **extra_format_kwargs: object):  # noqa: D103
    formatted_str = format_str.format(
        run_name=run_name,
        **_get_dist_config(strict=False),
        **extra_format_kwargs,
    )
    return formatted_str


format_name_with_dist.__doc__ = f"""
Format ``format_str`` with the ``run_name``, distributed variables, and ``extra_format_kwargs``.

The following format variables are available:

{FORMAT_NAME_WITH_DIST_TABLE}

For example, assume that the rank is ``0``. Then:

>>> from composer.utils import format_name_with_dist
>>> format_str = '{{run_name}}/rank{{rank}}.{{extension}}'
>>> format_name_with_dist(
...     format_str,
...     run_name='awesome_training_run',
...     extension='json',
... )
'awesome_training_run/rank0.json'

Args:
    format_str (str): The format string for the checkpoint filename.
    run_name (str): The value for the ``{{run_name}}`` format variable.
    extra_format_kwargs (object): Any additional :meth:`~str.format` kwargs.
"""

FORMAT_NAME_WITH_DIST_AND_TIME_TABLE = """
+----------------------------+------------------------------------------------------------+
| Variable                   | Description                                                |
+============================+============================================================+
| ``{run_name}``             | The name of the training run. See                          |
|                            | :attr:`.Logger.run_name`.                                  |
+----------------------------+------------------------------------------------------------+
| ``{rank}``                 | The global rank, as returned by                            |
|                            | :func:`~composer.utils.dist.get_global_rank`.              |
+----------------------------+------------------------------------------------------------+
| ``{local_rank}``           | The local rank of the process, as returned by              |
|                            | :func:`~composer.utils.dist.get_local_rank`.               |
+----------------------------+------------------------------------------------------------+
| ``{world_size}``           | The world size, as returned by                             |
|                            | :func:`~composer.utils.dist.get_world_size`.               |
+----------------------------+------------------------------------------------------------+
| ``{local_world_size}``     | The local world size, as returned by                       |
|                            | :func:`~composer.utils.dist.get_local_world_size`.         |
+----------------------------+------------------------------------------------------------+
| ``{node_rank}``            | The node rank, as returned by                              |
|                            | :func:`~composer.utils.dist.get_node_rank`.                |
+----------------------------+------------------------------------------------------------+
| ``{epoch}``                | The total epoch count, as returned by                      |
|                            | :meth:`~composer.core.time.Timestamp.epoch`.               |
+----------------------------+------------------------------------------------------------+
| ``{batch}``                | The total batch count, as returned by                      |
|                            | :meth:`~composer.core.time.Timestamp.batch`.               |
+----------------------------+------------------------------------------------------------+
| ``{batch_in_epoch}``       | The batch count in the current epoch, as returned by       |
|                            | :meth:`~composer.core.time.Timestamp.batch_in_epoch`.      |
+----------------------------+------------------------------------------------------------+
| ``{sample}``               | The total sample count, as returned by                     |
|                            | :meth:`~composer.core.time.Timestamp.sample`.              |
+----------------------------+------------------------------------------------------------+
| ``{sample_in_epoch}``      | The sample count in the current epoch, as returned by      |
|                            | :meth:`~composer.core.time.Timestamp.sample_in_epoch`.     |
+----------------------------+------------------------------------------------------------+
| ``{token}``                | The total token count, as returned by                      |
|                            | :meth:`~composer.core.time.Timestamp.token`.               |
+----------------------------+------------------------------------------------------------+
| ``{token_in_epoch}``       | The token count in the current epoch, as returned by       |
|                            | :meth:`~composer.core.time.Timestamp.token_in_epoch`.      |
+----------------------------+------------------------------------------------------------+
| ``{total_wct}``            | The total training duration in seconds, as returned by     |
|                            | :meth:`~composer.core.time.Timestamp.total_wct`.           |
+----------------------------+------------------------------------------------------------+
| ``{epoch_wct}``            | The epoch duration in seconds, as returned by              |
|                            | :meth:`~composer.core.time.Timestamp.epoch_wct`.           |
+----------------------------+------------------------------------------------------------+
| ``{batch_wct}``            | The batch duration in seconds, as returned by              |
|                            | :meth:`~composer.core.time.Timestamp.batch_wct`.           |
+----------------------------+------------------------------------------------------------+
"""


def format_name_with_dist_and_time(
    format_str: str,
    run_name: str,
    timestamp: Timestamp,
    **extra_format_kwargs: object,
):  # noqa: D103
    formatted_str = format_str.format(
        run_name=run_name,
        epoch=int(timestamp.epoch),
        batch=int(timestamp.batch),
        batch_in_epoch=int(timestamp.batch_in_epoch),
        sample=int(timestamp.sample),
        sample_in_epoch=int(timestamp.sample_in_epoch),
        token=int(timestamp.token),
        token_in_epoch=int(timestamp.token_in_epoch),
        total_wct=timestamp.total_wct.total_seconds(),
        epoch_wct=timestamp.epoch_wct.total_seconds(),
        batch_wct=timestamp.batch_wct.total_seconds(),
        **_get_dist_config(strict=False),
        **extra_format_kwargs,
    )
    return formatted_str


format_name_with_dist_and_time.__doc__ = f"""\
Format ``format_str`` with the ``run_name``, distributed variables, ``timestamp``, and ``extra_format_kwargs``.

In addition to the variables specified via ``extra_format_kwargs``, the following format variables are available:

{FORMAT_NAME_WITH_DIST_AND_TIME_TABLE}

For example, assume that the current epoch is ``0``, batch is ``0``, and rank is ``0``. Then:

>>> from composer.utils import format_name_with_dist_and_time
>>> format_str = '{{run_name}}/ep{{epoch}}-ba{{batch}}-rank{{rank}}.{{extension}}'
>>> format_name_with_dist_and_time(
...     format_str,
...     run_name='awesome_training_run',
...     timestamp=state.timestamp,
...     extension='json',
... )
'awesome_training_run/ep0-ba0-rank0.json'

Args:
    format_str (str): The format string for the checkpoint filename.
    run_name (str): The value for the ``{{run_name}}`` format variable.
    timestamp (Timestamp): The timestamp.
    extra_format_kwargs (object): Any additional :meth:`~str.format` kwargs.
"""


def parse_uri(uri: str) -> Tuple[str, str, str]:
    """Uses :py:func:`urllib.parse.urlparse` to parse the provided URI.

    Args:
        uri (str): The provided URI string

    Returns:
        Tuple[str, str, str]: A tuple containing the backend (e.g. s3), bucket name, and path.
                              Backend and bucket name will be empty string if the input is a local path
    """
    parse_result = urlparse(uri)
    backend, net_loc, path = parse_result.scheme, parse_result.netloc, parse_result.path
    bucket_name = net_loc if '@' not in net_loc else net_loc.split('@')[0]
    if backend == '' and bucket_name == '':
        return backend, bucket_name, path
    else:
        return backend, bucket_name, path.lstrip('/')


def maybe_create_object_store_from_uri(uri: str) -> Optional[ObjectStore]:
    """Automatically creates an :class:`composer.utils.ObjectStore` from supported URI formats.

    Currently supported backends are ``s3://``, ``oci://``, and local paths (in which case ``None`` will be returned)

    Args:
        uri (str): The path to (maybe) create an :class:`composer.utils.ObjectStore` from

    Raises:
        NotImplementedError: Raises when the URI format is not supported.

    Returns:
        Optional[ObjectStore]: Returns an :class:`composer.utils.ObjectStore` if the URI is of a supported format, otherwise None
    """
    backend, bucket_name, _ = parse_uri(uri)
    if backend == '':
        return None
    if backend == 's3':
        return S3ObjectStore(bucket=bucket_name)
    elif backend == 'wandb':
        raise NotImplementedError(f'There is no implementation for WandB load_object_store via URI. Please use '
                                  'WandBLogger')
    elif backend == 'gs':
        if 'GCS_KEY' not in os.environ or 'GCS_SECRET' not in os.environ:
            raise ValueError(
                'You must set the GCS_KEY and GCS_SECRET env variable with you HMAC access id and secret respectively')

        return LibcloudObjectStore(
            provider='google_storage',
            container=bucket_name,
            key_environ='GCS_KEY',  # Name of env variable for HMAC access id.
            secret_environ='GCS_SECRET',  # Name of env variable for HMAC secret.
        )
    elif backend == 'oci':
        return OCIObjectStore(bucket=bucket_name)
    else:
        raise NotImplementedError(f'There is no implementation for the cloud backend {backend} via URI. Please use '
                                  's3 or one of the supported object stores')


def maybe_create_remote_uploader_downloader_from_uri(
        uri: str, loggers: List[LoggerDestination]) -> Optional['RemoteUploaderDownloader']:
    """Automatically creates a :class:`composer.loggers.RemoteUploaderDownloader` from supported URI formats.

    Currently supported backends are ``s3://``, ``oci://``, and local paths (in which case ``None`` will be returned)

    Args:
        uri (str):The path to (maybe) create a :class:`composer.loggers.RemoteUploaderDownloader` from
        loggers (List[:class:`composer.loggers.LoggerDestination`]): List of the existing :class:`composer.loggers.LoggerDestination` s so as to not create a duplicate

    Raises:
        NotImplementedError: Raises when the URI format is not supported.

    Returns:
        Optional[RemoteUploaderDownloader]: Returns a :class:`composer.loggers.RemoteUploaderDownloader` if the URI is of a supported format, otherwise None
    """
    from composer.loggers import RemoteUploaderDownloader
    existing_remote_uds = [logger_dest for logger_dest in loggers if isinstance(logger_dest, RemoteUploaderDownloader)]
    backend, bucket_name, _ = parse_uri(uri)
    if backend == '':
        return None
    for existing_remote_ud in existing_remote_uds:
        if ((existing_remote_ud.remote_backend_name == backend) and
            (existing_remote_ud.remote_bucket_name == bucket_name)):
            warnings.warn(
                f'There already exists a RemoteUploaderDownloader object to handle the uri: {uri} you specified')
            return None
    if backend in ['s3', 'oci']:
        return RemoteUploaderDownloader(bucket_uri=f'{backend}://{bucket_name}')

    elif backend == 'gs':
        if 'GCS_KEY' not in os.environ or 'GCS_SECRET' not in os.environ:
            raise ValueError(
                'You must set the GCS_KEY and GCS_SECRET env variable with you HMAC access id and secret respectively')
        return RemoteUploaderDownloader(
            bucket_uri=f'libcloud://{bucket_name}',
            backend_kwargs={
                'provider': 'google_storage',
                'container': bucket_name,
                'key_environ': 'GCS_KEY',  # Name of env variable for HMAC access id.
                'secret_environ': 'GCS_SECRET',  # Name of env variable for HMAC secret.
            })

    elif backend == 'wandb':
        raise NotImplementedError(f'There is no implementation for WandB via URI. Please use '
                                  'WandBLogger with log_artifacts set to True')

    else:
        raise NotImplementedError(f'There is no implementation for the cloud backend {backend} via URI. Please use '
                                  's3 or one of the supported RemoteUploaderDownloader object stores')


def get_file(
    path: str,
    destination: str,
    object_store: Optional[Union[ObjectStore, LoggerDestination]] = None,
    overwrite: bool = False,
    progress_bar: bool = True,
):
    """Get a file from a local folder, URL, or object store.

    Args:
        path (str): The path to the file to retrieve.

            *   If ``object_store`` is specified, then the ``path`` should be the object name for the file to get.
                Do not include the the cloud provider or bucket name.

            *   If ``object_store`` is not specified but the ``path`` begins with ``http://`` or ``https://``,
                the object at this URL will be downloaded.

            *   If ``object_store`` is not specified, but the ``path`` begins with ``s3://``, or another backend
                supported by :meth:`composer.utils.maybe_create_object_store_from_uri` an appropriate object store
                will be created and used.

            *   Otherwise, ``path`` is presumed to be a local filepath.

        destination (str): The destination filepath.

            If ``path`` is a local filepath, then a symlink to ``path`` at ``destination`` will be created.
            Otherwise, ``path`` will be downloaded to a file at ``destination``.

        object_store (ObjectStore, optional): An :class:`~.ObjectStore`, if ``path`` is located inside
            an object store (i.e. AWS S3 or Google Cloud Storage). (default: ``None``)

            This :class:`~.ObjectStore` instance will be used to retrieve the file. The ``path`` parameter
            should be set to the object name within the object store.

            Set this parameter to ``None`` (the default) if ``path`` is a URL or a local file.

        overwrite (bool): Whether to overwrite an existing file at ``destination``. (default: ``False``)

        progress_bar (bool, optional): Whether to show a progress bar. Ignored if ``path`` is a local file.
            (default: ``True``)

    Raises:
        FileNotFoundError: If the ``path`` does not exist.
    """
    if object_store is None and not (path.lower().startswith('http://') or path.lower().startswith('https://')):
        object_store = maybe_create_object_store_from_uri(path)
        _, _, path = parse_uri(path)

    if path.endswith('.symlink'):
        with tempfile.TemporaryDirectory() as tmpdir:
            symlink_file_name = os.path.join(tmpdir, 'file.symlink')
            # Retrieve the symlink
            _get_file(
                path=path,
                destination=symlink_file_name,
                object_store=object_store,
                overwrite=False,
                progress_bar=progress_bar,
            )
            # Read object name in the symlink
            with open(symlink_file_name, 'r') as f:
                real_path = f.read()
                log.debug(f'Read path {real_path} from symlink file.')

        # Recurse
        return get_file(
            path=real_path,
            destination=destination,
            object_store=object_store,
            overwrite=overwrite,
            progress_bar=progress_bar,
        )

    try:
        _get_file(
            path=path,
            destination=destination,
            object_store=object_store,
            overwrite=overwrite,
            progress_bar=progress_bar,
        )
    except FileNotFoundError as e:
        new_path = path + '.symlink'
        try:
            # Follow the symlink
            return get_file(
                path=new_path,
                destination=destination,
                object_store=object_store,
                overwrite=overwrite,
                progress_bar=progress_bar,
            )
        except FileNotFoundError as ee:
            # Raise the original not found error first, which contains the path to the user-specified file
            raise e from ee


def _get_file(
    path: str,
    destination: str,
    object_store: Optional[Union[ObjectStore, LoggerDestination]],
    overwrite: bool,
    progress_bar: bool,
):
    # Underlying _get_file logic that does not deal with symlinks
    if object_store is not None:
        if isinstance(object_store, ObjectStore):
            total_size_in_bytes = object_store.get_object_size(path)
            object_store.download_object(
                object_name=path,
                filename=destination,
                callback=_get_callback(f'Downloading {path}') if progress_bar else None,
                overwrite=overwrite,
            )
        else:
            # Type LoggerDestination
            object_store.download_file(
                remote_file_name=path,
                destination=destination,
                progress_bar=progress_bar,
                overwrite=overwrite,
            )
        return

    if path.lower().startswith('http://') or path.lower().startswith('https://'):
        # it's a url
        with requests.get(path, stream=True) as r:
            try:
                r.raise_for_status()
            except requests.exceptions.HTTPError as e:
                if r.status_code == 404:
                    raise FileNotFoundError(f'URL {path} not found') from e
                raise e
            total_size_in_bytes = r.headers.get('content-length')
            if total_size_in_bytes is not None:
                total_size_in_bytes = int(total_size_in_bytes)
            else:
                total_size_in_bytes = 0

            tmp_path = destination + f'.{uuid.uuid4()}.tmp'
            try:
                with open(tmp_path, 'wb') as f:
                    for data in iterate_with_callback(
                            r.iter_content(2**20),
                            total_size_in_bytes,
                            callback=_get_callback(f'Downloading {path}') if progress_bar else None,
                    ):
                        f.write(data)
            except:
                # The download failed for some reason. Make a best-effort attempt to remove the temporary file.
                try:
                    os.remove(tmp_path)
                except OSError:
                    pass
                raise
            else:
                os.rename(tmp_path, destination)
        return

    # It's a local filepath
    if not os.path.exists(path):
        raise FileNotFoundError(f'Local path {path} does not exist')
    os.symlink(os.path.abspath(path), destination)


def _get_callback(description: str):
    if len(description) > 60:
        description = description[:42] + '...' + description[-15:]
    pbar = None

    def callback(num_bytes: int, total_size: int):
        nonlocal pbar
        if num_bytes == 0 or pbar is None:
            pbar = tqdm.tqdm(desc=description, total=total_size, unit='iB', unit_scale=True)
        n = num_bytes - pbar.n
        pbar.update(n)
        if num_bytes == total_size:
            pbar.close()

    return callback


def create_symlink_file(
    existing_path: str,
    destination_filename: Union[str, pathlib.Path],
):
    """Create a symlink file, which can be followed by :func:`get_file`.

    Unlike unix symlinks, symlink files can be created by this function are normal text files and can be
    uploaded to object stores via :meth:`.ObjectStore.upload_object` or loggers via :meth:`.Logger.upload_file`
    that otherwise would not support unix-style symlinks.

    Args:
        existing_path (str): The name of existing object that the symlink file should point to.
        destination_filename (str | pathlib.Path): The filename to which to write the symlink.
            It must end in ``'.symlink'``.
    """
    # Loggers might not natively support symlinks, so we emulate symlinks via text files ending with `.symlink`
    # This text file contains the name of the object it is pointing to.
    # Only symlink if we're uploading files to begin with
    # Write remote file name into file to emulate symlink
    # Add .symlink extension so we can identify as emulated symlink when downloading
    destination_filename = str(destination_filename)
    if not destination_filename.endswith('.symlink'):
        raise ValueError('The symlink filename must end with .symlink.')
    with open(destination_filename, 'x') as f:
        f.write(existing_path)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""The CIFAR ResNet torch module.

See the :doc:`Model Card </model_cards/resnet>` for more details.
"""

# Code below adapted from https://github.com/facebookresearch/open_lth
# and https://github.com/pytorch/vision

from typing import List, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision.models.resnet import BasicBlock

from composer.models import Initializer

__all__ = ['ResNetCIFAR', 'ResNet9']


class ResNetCIFAR(nn.Module):
    """A residual neural network as originally designed for CIFAR-10."""

    class Block(nn.Module):
        """A ResNet block."""

        def __init__(self, f_in: int, f_out: int, downsample: bool = False):
            super(ResNetCIFAR.Block, self).__init__()

            stride = 2 if downsample else 1
            self.conv1 = nn.Conv2d(f_in, f_out, kernel_size=3, stride=stride, padding=1, bias=False)
            self.bn1 = nn.BatchNorm2d(f_out)
            self.conv2 = nn.Conv2d(f_out, f_out, kernel_size=3, stride=1, padding=1, bias=False)
            self.bn2 = nn.BatchNorm2d(f_out)
            self.relu = nn.ReLU(inplace=True)

            # No parameters for shortcut connections.
            if downsample or f_in != f_out:
                self.shortcut = nn.Sequential(
                    nn.Conv2d(f_in, f_out, kernel_size=1, stride=2, bias=False),
                    nn.BatchNorm2d(f_out),
                )
            else:
                self.shortcut = nn.Sequential()

        def forward(self, x: torch.Tensor):
            out = self.relu(self.bn1(self.conv1(x)))
            out = self.bn2(self.conv2(out))
            out += self.shortcut(x)
            return self.relu(out)

    def __init__(self, plan: List[Tuple[int, int]], initializers: List[Initializer], outputs: int = 10):
        super(ResNetCIFAR, self).__init__()
        outputs = outputs or 10

        self.num_classes = outputs

        # Initial convolution.
        current_filters = plan[0][0]
        self.conv = nn.Conv2d(3, current_filters, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn = nn.BatchNorm2d(current_filters)
        self.relu = nn.ReLU(inplace=True)

        # The subsequent blocks of the ResNet.
        blocks = []
        for segment_index, (filters, num_blocks) in enumerate(plan):
            for block_index in range(num_blocks):
                downsample = segment_index > 0 and block_index == 0
                blocks.append(ResNetCIFAR.Block(current_filters, filters, downsample))
                current_filters = filters

        self.blocks = nn.Sequential(*blocks)

        # Final fc layer. Size = number of filters in last segment.
        self.fc = nn.Linear(plan[-1][0], outputs)
        self.criterion = nn.CrossEntropyLoss()

        for initializer in initializers:
            initializer = Initializer(initializer)
            self.apply(initializer.get_initializer())

    def forward(self, x: torch.Tensor):
        out = self.relu(self.bn(self.conv(x)))
        out = self.blocks(out)
        out = F.avg_pool2d(out, out.size()[3])
        out = out.view(out.size(0), -1)
        out = self.fc(out)
        return out

    @staticmethod
    def is_valid_model_name(model_name: str):
        valid_model_names = [f'resnet_{layers}' for layers in (20, 56)]
        return (model_name in valid_model_names)

    @staticmethod
    def get_model_from_name(model_name: str, initializers: List[Initializer], outputs: int = 10):
        """The naming scheme for a ResNet is ``'resnet_D[_W]'``.

        D is the model depth (e.g. ``'resnet_56'``)
        """

        if not ResNetCIFAR.is_valid_model_name(model_name):
            raise ValueError('Invalid model name: {}'.format(model_name))

        depth = int(model_name.split('_')[-1])  # for resnet56, depth 56, width 16
        if len(model_name.split('_')) == 2:
            width = 16
        else:
            width = int(model_name.split('_')[3])

        if (depth - 2) % 3 != 0:
            raise ValueError('Invalid ResNetCIFAR depth: {}'.format(depth))
        num_blocks = (depth - 2) // 6

        model_arch = {
            56: [(width, num_blocks), (2 * width, num_blocks), (4 * width, num_blocks)],
            20: [(width, num_blocks), (2 * width, num_blocks), (4 * width, num_blocks)],
        }

        return ResNetCIFAR(model_arch[depth], initializers, outputs)


# adapted from https://raw.githubusercontent.com/matthias-wright/cifar10-resnet/master/model.py
# under the MIT license
class ResNet9(nn.Module):
    """A 9-layer residual network, excluding BatchNorms and activation functions.

    Based on the myrtle.ai `blog`_ and Deep Residual Learning for Image Recognition (`He et al, 2015`_).

    Args:
        num_classes (int, optional): The number of classes. Needed for classification tasks. Default: ``10``.

    .. _blog: https://myrtle.ai/learn/how-to-train-your-resnet-4-architecture/
    .. _He et al, 2015: https://arxiv.org/abs/1512.03385
    """

    def __init__(self, num_classes: int = 10):
        super().__init__()

        self.body = nn.Sequential(
            nn.Conv2d(in_channels=3, out_channels=64, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(num_features=64, momentum=0.9),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels=64, out_channels=128, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(num_features=128, momentum=0.9),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),
            BasicBlock(inplanes=128, planes=128, stride=1),
            nn.Conv2d(in_channels=128, out_channels=256, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(num_features=256, momentum=0.9),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),
            nn.Conv2d(in_channels=256, out_channels=256, kernel_size=3, stride=1, padding=1, bias=False),
            nn.BatchNorm2d(num_features=256, momentum=0.9),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),
            BasicBlock(inplanes=256, planes=256, stride=1),
        )

        self.fc = nn.Linear(in_features=256, out_features=num_classes, bias=True)

    def forward(self, x):
        out = self.body(x)
        out = F.avg_pool2d(out, out.size()[3])
        out = out.view(out.size(0), -1)
        out = self.fc(out)
        return out



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Helpers to get items and set items in a batch."""

from operator import attrgetter, itemgetter
from typing import Any, Callable, Sequence, Union, cast

__all__ = ['batch_get', 'batch_set']


def batch_get(batch: Any, key: Union[str, int, Callable, Any]):
    """Indexes into the batch given the key.

    >>> from composer.utils.batch_helpers import batch_get
    >>> batch_get([1,2,3], 1)
    2
    >>> batch_get({'a':1, 'b':7}, 'b')
    7
    >>> batch_get([{'a':1, 'b':7},{'c':5}], lambda x: x[1]['c'])
    5
    >>> batch_get([{'a':1, 'b':7},{'c':5}], (lambda x: x[1]['c'], lambda x: 10))
    5

    Args:
        batch (Any): An object that contains the input and label of the items in the batch.
            Can be any abritrary type that user creates, but we assume some sort of
            sequence (list, tuple, tensor, array), mapping (dictionary),
            or attribute store (object with data members, namedtuple).
        key (str | int | Tuple[Callable, Callable] | Any, optional): A key to index into the batch or a
                user-specified function to do the extracting. A pair of callables is also
                supported for cases where a get and set function pair are both passed
                (like in Algorithms). The getter is assumed to be the first of the pair.

    Returns:
        The part of the batch specified by the key. This could be any type
            depending on what the batch is composed of.
    """
    # Case 1: key is a tuple of (getter, setter).
    if (isinstance(key, Sequence) and not isinstance(key, str) and _is_key_get_and_set_fn_pair(key)):
        get_fn, _ = key
        return get_fn(batch)

    # Case 2: key is a getter Callable.
    if isinstance(key, Callable):
        return key(batch)

    # Case 3: key some sort of index or key to use to directly extract from the batch.
    try:
        return itemgetter(key)(batch)
    except (IndexError, TypeError):
        try:
            return itemgetter(*key)(batch)
        except TypeError:
            try:
                return attrgetter(cast(str, key))(batch)
            except TypeError:
                return attrgetter(*key)(batch)


def batch_set(batch: Any, key: Union[str, int, Callable, Any], value: Any) -> Any:
    """Indexes into the batch given the key and sets the element at that index to value.

    This is not an in-place operation for batches of type tuple as tuples are not mutable.

    >>> from composer.utils.batch_helpers import batch_set
    >>> batch_set([1,2,3], key=1, value=8)
    [1, 8, 3]
    >>> batch_set({'a':1, 'b':7}, key='b', value=11)
    {'a': 1, 'b': 11}
    >>> def setter(batch, value):
    ...     batch[1]['d'] = value
    ...     return batch
    ...
    >>> batch_set([{'a':1, 'b':7},{'d':3}], key=setter, value=20)
    [{'a': 1, 'b': 7}, {'d': 20}]
    >>> batch_set([{'a':1, 'b':7},{'d':3}], key=(lambda x: x[0]['b'], setter), value=20)
    [{'a': 1, 'b': 7}, {'d': 20}]


    Args:
        batch (Any): An object that contains the input and label of the items in the batch.
            Can be any abritrary type that user creates, but we assume some sort of
            sequence (list, tuple, tensor, array), mapping (dictionary),
            or attribute store (object with data members, namedtuple).
        key (str | int | Tuple[Callable, Callable] | Any, optional): A key to index into the batch or a user-specified function
            to do the setting. A pair of callables is also supported for cases where a get
            and set function pair are both passed (like in Algorithms). The setter is
            assumed to be the second of the pair.
        value (Any): The value that batch[key] or batch.key gets set to.

    Returns:
        batch (Any): updated batch with value set at key.

    """
    # Case 1: key is a tuple of (getter, setter) callables.
    if (isinstance(key, Sequence) and not isinstance(key, str) and _is_key_get_and_set_fn_pair(key)):
        _, set_fn = key
        return set_fn(batch, value)

    # Case 2: key is a callable.
    if isinstance(key, Callable):
        return key(batch, value)

    # Case 4: key is sequence of sub-keys.
    if isinstance(key, Sequence) and not isinstance(key, str):
        return _batch_set_multiple(batch, key, value)

    # Case 5: key is single object, like string or int.
    else:
        return _batch_set(batch, key, value)


def _batch_set(batch: Any, key: Any, value: Any) -> Any:
    """Sets a key value pair in a non-tuple batch."""
    if isinstance(batch, tuple):
        return _batch_set_tuple(batch, key, value)

    try:
        # Check if one can do a __getitem__ before doing a __setitem__ because dicts can
        # do __setitem__ for elements not in the dict and we do not want that.
        # Note for defaultdict and Counter objects, just calling batch[key] for
        # with a new keyword will create a new key, value pair in the object.
        batch[key]
        batch[key] = value

    # The only acceptable TypeErrors are for an object that doesn't have a __setitem__ or a __getitem__,
    # which is TypeError("... object does not support item assignment") and TypeError('.. object is not subscriptable')
    except TypeError as e:
        if 'object does not support item assignment' in str(e) or 'object is not subscriptable' in str(e):
            pass
        else:  # Other type errors should be raised.
            raise e
    else:
        return batch

    try:
        # Make sure batch has key before setting it.
        getattr(batch, key)
        setattr(batch, key, value)

    # If both (setattr or getattr) and __setitem__ raise exceptions then raise both of them.
    except (AttributeError, TypeError) as e:
        raise RuntimeError(
            f'Unable to set key {key} to value {value} on batch {batch}. Please specify a custom set_fn, if necessary.')
    else:
        return batch


def _batch_set_multiple(batch: Any, key: Any, value: Any) -> Any:
    """Sets multiple key value pairs in a non-tuple batch."""
    # Numpy arrays and Torch tensors can take tuples and lists as keys, so try to do a normal
    # __getitem__ call before resulting to list comprehension.
    try:
        # Check if one can do a __getitem__ before doing a __setitem__ because dicts can
        # do __setitem__ for elements not in the dict and we do not want that.
        batch[key]
        batch[key] = value
        return batch
    # Indexing a list with a sequence results in TypeError
    # Indexing an array/tensor with a sequence that is longer than the rank of the array
    # results in an IndexError.
    except (IndexError, TypeError, KeyError):
        pass
    if not hasattr(value, '__len__') or isinstance(value, str):
        raise ValueError(f'value must be a sequence or array or tensor! and not {type(value)}')
    if len(key) != len(value):
        raise ValueError(f'value must be the same length as key ({len(key)}), but it is {len(value)} instead')
    for single_key, single_value in zip(key, value):
        batch = _batch_set(batch, single_key, single_value)
    return batch


def _batch_set_tuple(batch: Any, key: Union[int, str], value: Any) -> Any:
    """Sets key value pairs in tuples and NamedTuples."""
    if hasattr(batch, '_fields'):  # NamedTuple
        if isinstance(key, str):
            batch = batch._replace(**{key: value})
        else:
            batch_list = list(batch)
            batch_list[key] = value
            batch = batch._make(batch_list)

    else:  # Normal tuple.
        batch = list(batch)
        batch[key] = value
        batch = tuple(batch)

    return batch


def _is_key_get_and_set_fn_pair(key):
    if all([callable(key_element) for key_element in key]):
        if len(key) == 2:
            return True
        else:
            raise ValueError(f"If key is a sequence of Callables, it should be of length 2' not {len(key)}")
    return False



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""`BlurPool <http://proceedings.mlr.press/v97/zhang19a.html>`_ adds anti-aliasing filters to convolutional layers to
increase accuracy and invariance to small shifts in the input.

See :class:`~composer.algorithms.BlurPool` or the :doc:`Method Card </method_cards/blurpool>` for details.
"""

from composer.algorithms.blurpool.blurpool import BlurPool as BlurPool
from composer.algorithms.blurpool.blurpool import apply_blurpool as apply_blurpool
from composer.algorithms.blurpool.blurpool_layers import BlurConv2d as BlurConv2d
from composer.algorithms.blurpool.blurpool_layers import BlurMaxPool2d as BlurMaxPool2d
from composer.algorithms.blurpool.blurpool_layers import BlurPool2d as BlurPool2d
from composer.algorithms.blurpool.blurpool_layers import blur_2d as blur_2d
from composer.algorithms.blurpool.blurpool_layers import blurmax_pool2d as blurmax_pool2d

__all__ = [
    'BlurPool',
    'apply_blurpool',
    'BlurConv2d',
    'BlurMaxPool2d',
    'BlurPool2d',
    'blur_2d',
    'blurmax_pool2d',
]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Merge trace files together.

To run:

.. code-block::

    python -m composer.profiler.json_trace_merger -o merged_trace_output.json path/to/input_file_1.json path/to/input_file_2.json ...

To view the traces, open a Google Chrome browser window, navigate to ``chrome://tracing`` and load the ``merged_trace_output.json``
to visualize the trace.
"""
import argparse
import json
import pathlib
from typing import Dict, List, Tuple, Union

__all__ = ['merge_traces']


def _load_trace(file: Union[str, pathlib.Path]) -> Union[Dict, List]:
    with open(file, 'r') as f:
        trace_str = f.read().strip()
        if trace_str.startswith('['):
            if trace_str.endswith('}'):
                trace_str += ']'
            if trace_str.endswith(','):
                trace_str = trace_str[-1] + ']'
        return json.loads(trace_str)


def _get_global_rank_from_file(file: Union[str, pathlib.Path]) -> int:
    trace_json = _load_trace(file)
    if isinstance(trace_json, list):
        for event in trace_json:
            if event['ph'] == 'M' and event['name'] == 'global_rank':
                return event['args']['value']
    else:
        assert isinstance(trace_json, dict)
        return trace_json['global_rank']
    raise RuntimeError('global rank not found in file')


def _get_rank_to_clock_syncs(trace_files: Tuple[Union[str, pathlib.Path], ...]) -> Dict[int, int]:
    rank_to_clock_sync: Dict[int, int] = {}
    for filename in trace_files:
        rank = _get_global_rank_from_file(filename)
        trace_json = _load_trace(filename)
        if isinstance(trace_json, list):
            for event in trace_json:
                if event['ph'] == 'M' and event['name'] == 'clock_sync_timestamp_us':
                    clock_sync = event['args']['value']
                    rank_to_clock_sync[rank] = clock_sync
                    break
        else:
            assert isinstance(trace_json, dict)
            if trace_json.get('clock_sync_timestamp_us') is not None:
                rank_to_clock_sync[rank] = trace_json['clock_sync_timestamp_us']

    return rank_to_clock_sync


def merge_traces(output_file: Union[str, pathlib.Path], *trace_files: Union[str, pathlib.Path]):
    """Merge profiler output JSON trace files together.

    This function will update the trace events such that:

    - The ``pid`` will be set to the global rank.
    - The ``ts`` is synchronized with that of the rank 0 process.
    - The backward pass process appears below the forward process.

    Args:
        output_file (str | pathlib.Path): The file to write the merged trace to
        trace_files (str | pathlib.Path): Variable number of trace files to merge together
    """
    ranks_to_clock_sync = _get_rank_to_clock_syncs(trace_files)
    rank_to_backwards_thread = {}
    rank_to_seen_threads = {rank: set() for rank in ranks_to_clock_sync.keys()}

    # Local rank zero will be the lowest global rank
    # Use that as the base timestamp for clock syncing
    lowest_rank = float('inf')
    for k in ranks_to_clock_sync:
        lowest_rank = min(k, lowest_rank)
    assert isinstance(lowest_rank, int), 'there should be at least one rank'
    rank_zero_clock_sync = ranks_to_clock_sync[lowest_rank]

    with open(output_file, 'w+') as output_f:
        is_first_line = True
        output_f.write('[')
        for trace_filename in trace_files:
            rank = _get_global_rank_from_file(trace_filename)
            clock_sync_diff = rank_zero_clock_sync - ranks_to_clock_sync[rank]

            with open(trace_filename, 'r') as trace_f:
                trace_data = json.load(trace_f)

                if isinstance(trace_data, list):
                    trace_list = trace_data
                else:
                    assert isinstance(trace_data, dict)
                    trace_list = trace_data['traceEvents']

                for event in trace_list:
                    if 'pid' not in event:
                        # we need the pid to merge
                        continue
                    if 'tid' not in event:
                        continue
                    if 'PyTorch Profiler' in str(event['tid']):
                        # skip this line; it pollutes the UI
                        continue
                    if 'ts' in event:
                        event['ts'] = event['ts'] + clock_sync_diff
                    event['pid'] = rank
                    if event['tid'] not in rank_to_seen_threads[rank]:
                        # By default, make all threads display last
                        # The training loop thread later sets itself as thread 0
                        # and the backwards pass thread is set as thread 1
                        if not is_first_line:
                            output_f.write(',')
                        output_f.write('\n    ')
                        json.dump(
                            {
                                'name': 'thread_sort_index',
                                'ph': 'M',
                                'pid': rank,
                                'tid': event['tid'],
                                'args': {
                                    'sort_index': 99999,
                                }
                            }, output_f)
                        rank_to_seen_threads[rank].add(event['tid'])
                        is_first_line = False
                    if event['name'] == 'MulBackward0':
                        rank_to_backwards_thread[rank] = event['tid']
                    if not is_first_line:
                        output_f.write(',')
                    is_first_line = False
                    output_f.write(f'\n    ')
                    json.dump(event, output_f)
        for pid, tid in rank_to_backwards_thread.items():
            output_f.write(',\n    ')
            json.dump({
                'name': 'thread_sort_index',
                'ph': 'M',
                'pid': pid,
                'tid': tid,
                'args': {
                    'sort_index': 1
                }
            }, output_f)

        output_f.write('\n]\n')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('input_files', type=str, nargs='+', help='Input files')
    parser.add_argument('-o', '--output_file', help='Output File', required=True)
    args = parser.parse_args()
    output_file = args.output_file
    input_files = args.input_files
    merge_traces(output_file, *input_files)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Utilities to track training progress in terms of epochs, batches, samples, and tokens.

Callbacks, algorithms, and schedulers can use the current training time to fire at certain points in the training
process.

The :class:`~.time.Timestamp` class tracks the total number of epochs, batches, samples, and tokens. The trainer is
responsible for updating it at the end of every epoch and batch.  There is only one instance of the
:class:`~.time.Timestamp`, which is attached to the :class:`~.state.State`.

The :class:`~.time.Time` class represents static durations of training time or points in the training process in terms
of a specific :class:`~.time.TimeUnit` enum. This class supports comparisons, arithmetic, and conversions.

See the :doc:`Time Guide </trainer/time>` for more details on tracking time during training.
"""
from __future__ import annotations

import datetime
import re
from typing import Any, Dict, Generic, Optional, TypeVar, Union, cast

from composer.core.serializable import Serializable
from composer.utils import StringEnum

__all__ = ['TimeUnit', 'Time', 'Timestamp', 'ensure_time']


class TimeUnit(StringEnum):
    """Enum class to represent units of time for the training process.

    Attributes:
        EPOCH (str): Epochs.
        BATCH (str): Batches (i.e. number of optimization steps)
        SAMPLE (str): Samples.
        TOKEN (str): Tokens. Applicable for natural language processing (NLP) models.
        DURATION (str): Fraction of the training process complete, on ``[0.0, 1.0)``
    """
    EPOCH = 'ep'
    BATCH = 'ba'
    SAMPLE = 'sp'
    TOKEN = 'tok'
    DURATION = 'dur'


# regex for parsing integers / decimals / scientific notation
_NUM_REGEX = r'-?[\d.]+(?:e-?\d+)?'

# regex for parsing a time string.
_TIME_STR_REGEX = re.compile(r'^(?:' + r'|'.join(fr'(?:({_NUM_REGEX})({time_unit.value}))' for time_unit in TimeUnit) +
                             r')$',
                             flags=re.IGNORECASE)

TValue = TypeVar('TValue', int, float)


class Time(Generic[TValue], Serializable):
    """Time represents static durations of training time in terms of a :class:`TimeUnit` enum.

    See the :doc:`Time Guide </trainer/time>` for more details on tracking time during training.

    To construct an instance of :class:`Time`, you can either:

    #. Use a value followed by a :class:`TimeUnit` enum or string. For example,

    >>> Time(5, TimeUnit.EPOCH)  # describes 5 epochs.
    Time(5, TimeUnit.EPOCH)
    >>> Time(30_000, "tok")  # describes 30,000 tokens.
    Time(30000, TimeUnit.TOKEN)
    >>> Time(0.5, "dur")  # describes 50% of the training process.
    Time(0.5, TimeUnit.DURATION)

    #. Use one of the helper methods. See:

    - :meth:`Time.from_epoch`
    - :meth:`Time.from_batch`
    - :meth:`Time.from_sample`
    - :meth:`Time.from_token`
    - :meth:`Time.from_duration`
    - :meth:`Time.from_timestring`.

    :class:`Time` supports addition and subtraction with other :class:`Time` instances that share the same
    :class:`TimeUnit`. For example:

    >>> Time(1, TimeUnit.EPOCH) + Time(2, TimeUnit.EPOCH)
    Time(3, TimeUnit.EPOCH)

    :class:`Time` supports multiplication. The multiplier must be either a number or have units of
    :attr:`TimeUnit.DURATION`. The multiplicand is scaled, and its units are kept.

    >>> Time(2, TimeUnit.EPOCH) * 0.5
    Time(1, TimeUnit.EPOCH)

    >>> Time(2, TimeUnit.EPOCH) * Time(0.5, TimeUnit.DURATION)
    Time(1, TimeUnit.EPOCH)


    :class:`Time` supports division. If the divisor is an instance of :class:`Time`, then it
    must have the same units as the dividend, and the result has units of :attr:`TimeUnit.DURATION`.
    For example:

    >>> Time(4, TimeUnit.EPOCH) / Time(2, TimeUnit.EPOCH)
    Time(2.0, TimeUnit.DURATION)

    If the divisor is number, then the dividend is scaled, and it keeps its units. For example:

    >>> Time(4, TimeUnit.EPOCH) / 2
    Time(2, TimeUnit.EPOCH)

    Args:
        value (int | float): The amount of time.
        unit (str | TimeUnit): The :class:`TimeUnit` for ``value``.
    """

    def __init__(
        self,
        value: TValue,
        unit: Union[str, TimeUnit],
    ):
        unit = TimeUnit(unit)
        if unit == TimeUnit.DURATION:
            value = cast(TValue, float(value))
        else:
            if not isinstance(value, int):
                raise TypeError(f'value {value} is of type {type(value)}. Units {unit} require integer values.')
        self._value, self._unit = value, TimeUnit(unit)

    @classmethod
    def from_epoch(cls, epoch: int) -> Time:
        """Create a :class:`Time` with units of :attr:`TimeUnit.EPOCH`.

        Equivalent to ``Time(epoch, TimeUnit.EPOCH)``.

        Args:
            epoch (int): Number of epochs.

        Returns:
            Time: :class:`Time` instance, in epochs.
        """
        return cls(epoch, TimeUnit.EPOCH)

    @classmethod
    def from_batch(cls, batch: int) -> Time:
        """Create a :class:`Time` with units of :attr:`TimeUnit.BATCH`.

        Equivalent to ``Time(batch, TimeUnit.BATCH)``.

        Args:
            batch (int): Number of batches.

        Returns:
            Time: :class:`Time` instance, in batches.
        """
        return cls(batch, TimeUnit.BATCH)

    @classmethod
    def from_sample(cls, sample: int) -> Time:
        """Create a :class:`Time` with units of :attr:`TimeUnit.SAMPLE`.

        Equivalent to ``Time(sample, TimeUnit.SAMPLE)``.

        Args:
            sample (int): Number of samples.

        Returns:
            Time: :class:`Time` instance, in samples.
        """
        return cls(sample, TimeUnit.SAMPLE)

    @classmethod
    def from_token(cls, token: int) -> Time:
        """Create a :class:`Time` with units of :attr:`TimeUnit.TOKEN`.

        Equivalent to ``Time(sample, TimeUnit.TOKEN)``.

        Args:
            token (int): Number of tokens.

        Returns:
            Time: :class:`Time` instance, in tokens.
        """
        return cls(token, TimeUnit.TOKEN)

    @classmethod
    def from_duration(cls, duration: float) -> Time:
        """Create a :class:`Time` with units of :attr:`TimeUnit.DURATION`.

        Equivalent to ``Time(duration, TimeUnit.DURATION)``.

        Args:
            duration (float): Duration of the training process. Should be on ``[0, 1)``
                where ``0`` represents the beginning of the training process and ``1``
                represents a completed training process.

        Returns:
            Time: :class:`Time` instance, in duration.
        """
        return cls(duration, TimeUnit.DURATION)

    @property
    def value(self) -> TValue:
        """The value of the time, as a number."""
        return self._value

    @property
    def unit(self) -> TimeUnit:
        """The unit of the time."""
        return self._unit

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.value}, {self.unit})'

    def __str__(self) -> str:
        return f'{self.value}{self.unit.value}'

    def to_timestring(self):
        """Get the time-string representation.

        For example:

        >>> Time(5, TimeUnit.EPOCH).to_timestring()
        '5ep'

        Returns:
            str: The time-string representation.
        """
        return str(self)

    def _parse(self, other: object) -> Time:
        # parse ``other`` into a Time object
        if isinstance(other, Time):
            return other
        if isinstance(other, int):
            return Time(other, self.unit)
        if isinstance(other, str):
            other_parsed = Time.from_timestring(other)
            return other_parsed

        raise TypeError(f'Cannot convert type {other} to {self.__class__.__name__}')

    def _cmp(self, other: Union[int, float, Time, str]) -> int:
        # When doing comparisions, and other is an integer (or float), we can safely infer
        # the unit from self.unit
        # E.g. calls like this should be allowed: if batch < 42: do_something()
        # This eliminates the need to call .value everywhere
        if not isinstance(other, (int, float, Time, str)):
            return NotImplemented
        if isinstance(other, (int, float)):
            other = type(self)(other, self.unit)
        other = self._parse(other)
        if self.unit != other.unit:
            raise RuntimeError(f'Cannot compare {self} to {other} since they have different units.')
        if self.value < other.value:
            return -1
        if self.value == other.value:
            return 0
        assert self.value > other.value
        return 1

    def __eq__(self, other: Union[int, float, Time, str]):
        return self._cmp(other) == 0

    def __ne__(self, other: Union[int, float, Time, str]):
        return self._cmp(other) != 0

    def __lt__(self, other: Union[int, float, Time, str]):
        return self._cmp(other) < 0

    def __le__(self, other: Union[int, float, Time, str]):
        return self._cmp(other) <= 0

    def __gt__(self, other: Union[int, float, Time, str]):
        return self._cmp(other) > 0

    def __ge__(self, other: Union[int, float, Time, str]):
        return self._cmp(other) >= 0

    def __add__(self, other: Union[int, float, Time, str]) -> Time[TValue]:
        other = self._parse(other)
        if self.unit != other.unit:
            raise RuntimeError(f'Cannot add {self} to {other} since they have different units.')
        return Time(self.value + other.value, self.unit)

    def __radd__(self, other: Union[int, float, Time, str]) -> Time[TValue]:
        return self + other

    def __sub__(self, other: Union[int, float, Time, str]) -> Time[TValue]:
        other = self._parse(other)
        if self.unit != other.unit:
            raise RuntimeError(f'Cannot subtract {other} from {self} since they have different units.')
        return Time(self.value - other.value, self.unit)

    def __rsub__(self, other: Union[int, float, Time, str]) -> Time[TValue]:
        return (-self) + other

    def __neg__(self) -> Time[TValue]:
        return Time(cast(TValue, -self.value), self.unit)

    def __pos__(self) -> Time[TValue]:
        return Time(self.value, self.unit)

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __truediv__(self, other: object) -> Time[float]:
        if isinstance(other, (float, int)):
            return Time(type(self.value)(self.value / other), self.unit)
        other = self._parse(other)
        if self.unit != other.unit:
            raise RuntimeError(f'Cannot divide {self} by {other} since they have different units.')
        return Time(self.value / other.value, TimeUnit.DURATION)

    def __mul__(self, other: object):
        if isinstance(other, (float, int)):
            # Scale by the value.
            return Time(type(self.value)(self.value * other), self.unit)
        other = self._parse(other)
        if other.unit != TimeUnit.DURATION and self.unit != TimeUnit.DURATION:
            raise RuntimeError(f'Multiplication is supported only if one of the units is Duration')
        real_unit = self.unit if other.unit == TimeUnit.DURATION else other.unit
        real_type = float if real_unit == TimeUnit.DURATION else int
        return Time(real_type(self.value * other.value), real_unit)

    def __rmul__(self, other: object):
        return self * other

    def __hash__(self):
        return hash((self.value, self.unit))

    @classmethod
    def from_timestring(cls, timestring: str) -> Time:
        """Parse a time string into a :class:`Time` instance.

        A time string is a numerical value followed by the value of a :class:`TimeUnit` enum. For example:

        >>> Time.from_timestring("5ep")  # describes 5 epochs.
        Time(5, TimeUnit.EPOCH)
        >>> Time.from_timestring("3e4tok")  # describes 30,000 tokens.
        Time(30000, TimeUnit.TOKEN)
        >>> Time.from_timestring("0.5dur")  # describes 50% of the training process.
        Time(0.5, TimeUnit.DURATION)

        Returns:
            Time: An instance of :class:`Time`.
        """
        match = _TIME_STR_REGEX.findall(timestring)
        if len(match) != 1:
            raise ValueError(f'Invalid time string: {timestring}')
        match = match[0]
        match = [x for x in match if x != '']
        assert len(match) == 2, 'each match should have a number followed by the key'
        value = match[0]
        unit = TimeUnit(match[1])
        value = float(value)  # always parsing first as float b/c it could be scientific notation
        if unit != TimeUnit.DURATION:
            if int(value) != value:
                raise TypeError(f'value {value} is not an integer. Units {unit} require integer values.')
            value = int(value)
        return cls(value, unit)


class Timestamp(Serializable):
    """Timestamp represents a snapshot of the current training progress.

    The timestamp measures training progress in terms of epochs, batches, samples, tokens, and wall clock time.
    Timestamps are not updated in-place.

    See the :doc:`Time Guide </trainer/time>` for more details on tracking time during training.

    Args:
        epoch (int | Time[int], optional): The epoch.
        batch (int | Time[int], optional): the batch.
        sample (int | Time[int], optional): The sample.
        token (int | Time[int], optional): The token.
        batch_in_epoch (int | Time[int], optional): The batch in the epoch.
        sample_in_epoch (int | Time[int], optional): The sample in the epoch.
        token_in_epoch (int | Time[int], optional): The token in the epoch.
        total_wct (datetime.timedelta, optional): The total wall-clock duration.
        epoch_wct (datetime.timedelta, optional): The wall-clock duration of the last epoch.
        batch_wct (datetime.timedelta, optional): The wall-clock duration of the last batch.
    """

    def __init__(
        self,
        epoch: Union[int, Time[int]] = 0,
        batch: Union[int, Time[int]] = 0,
        sample: Union[int, Time[int]] = 0,
        token: Union[int, Time[int]] = 0,
        batch_in_epoch: Union[int, Time[int]] = 0,
        sample_in_epoch: Union[int, Time[int]] = 0,
        token_in_epoch: Union[int, Time[int]] = 0,
        total_wct: Optional[datetime.timedelta] = None,
        epoch_wct: Optional[datetime.timedelta] = None,
        batch_wct: Optional[datetime.timedelta] = None,
    ):
        epoch = ensure_time(epoch, TimeUnit.EPOCH)
        if epoch.unit != TimeUnit.EPOCH:
            raise ValueError(f'The `epoch` argument has units of {epoch.unit}; not {TimeUnit.EPOCH}.')
        self._epoch = epoch

        batch = ensure_time(batch, TimeUnit.BATCH)
        if batch.unit != TimeUnit.BATCH:
            raise ValueError(f'The `batch` argument has units of {batch.unit}; not {TimeUnit.BATCH}.')
        self._batch = batch

        sample = ensure_time(sample, TimeUnit.SAMPLE)
        if sample.unit != TimeUnit.SAMPLE:
            raise ValueError(f'The `sample` argument has units of {sample.unit}; not {TimeUnit.SAMPLE}.')
        self._sample = sample

        token = ensure_time(token, TimeUnit.TOKEN)
        if token.unit != TimeUnit.TOKEN:
            raise ValueError(f'The `token` argument has units of {token.unit}; not {TimeUnit.TOKEN}.')
        self._token = token

        batch_in_epoch = ensure_time(batch_in_epoch, TimeUnit.BATCH)
        if batch_in_epoch.unit != TimeUnit.BATCH:
            raise ValueError((f'The `batch_in_epoch` argument has units of {batch_in_epoch.unit}; '
                              f'not {TimeUnit.BATCH}.'))
        self._batch_in_epoch = batch_in_epoch

        sample_in_epoch = ensure_time(sample_in_epoch, TimeUnit.SAMPLE)
        if sample_in_epoch.unit != TimeUnit.SAMPLE:
            raise ValueError((f'The `sample_in_epoch` argument has units of {sample_in_epoch.unit}; '
                              f'not {TimeUnit.SAMPLE}.'))
        self._sample_in_epoch = sample_in_epoch

        token_in_epoch = ensure_time(token_in_epoch, TimeUnit.TOKEN)
        if token_in_epoch.unit != TimeUnit.TOKEN:
            raise ValueError((f'The `token_in_epoch` argument has units of {token_in_epoch.unit}; '
                              f'not {TimeUnit.TOKEN}.'))
        self._token_in_epoch = token_in_epoch

        if total_wct is None:
            total_wct = datetime.timedelta(seconds=0)
        self._total_wct = total_wct

        if epoch_wct is None:
            epoch_wct = datetime.timedelta(seconds=0)
        self._epoch_wct = epoch_wct

        if batch_wct is None:
            batch_wct = datetime.timedelta(seconds=0)
        self._batch_wct = batch_wct

    def state_dict(self) -> Dict[str, Any]:
        return {
            'epoch': self.epoch.value,
            'batch': self.batch.value,
            'sample': self.sample.value,
            'token': self.token.value,
            'batch_in_epoch': self.batch_in_epoch.value,
            'sample_in_epoch': self.sample_in_epoch.value,
            'token_in_epoch': self.token_in_epoch.value,
            'total_wct': self.total_wct,
            'epoch_wct': self.epoch_wct,
            'batch_wct': self.batch_wct,
        }

    def get_state(self) -> Dict[str, Union[Time[int], datetime.timedelta]]:
        """Returns all values of the timestamp object in a dictionary.

        Returns:
            Dict[str, Union[Time[int], datetime.timedelta]]: All values of the timestamp object.
        """
        return {
            'epoch': self.epoch,
            'batch': self.batch,
            'sample': self.sample,
            'token': self.token,
            'batch_in_epoch': self.batch_in_epoch,
            'sample_in_epoch': self.sample_in_epoch,
            'token_in_epoch': self.token_in_epoch,
            'total_wct': self.total_wct,
            'epoch_wct': self.epoch_wct,
            'batch_wct': self.batch_wct,
        }

    def load_state_dict(self, state: Dict[str, Any]) -> None:
        self._epoch = Time(state['epoch'], TimeUnit.EPOCH)
        self._batch = Time(state['batch'], TimeUnit.BATCH)
        self._sample = Time(state['sample'], TimeUnit.SAMPLE)
        self._token = Time(state['token'], TimeUnit.TOKEN)
        self._batch_in_epoch = Time(state['batch_in_epoch'], TimeUnit.BATCH)
        self._sample_in_epoch = Time(state['sample_in_epoch'], TimeUnit.SAMPLE)
        self._token_in_epoch = Time(state['token_in_epoch'], TimeUnit.TOKEN)
        # Wall clock time tracking was added in composer v0.7.0
        # Using conditional checks as not to break old checkpoints
        if 'total_wct' in state:
            self._total_wct = state['total_wct']
        if 'epoch_wct' in state:
            self._epoch_wct = state['epoch_wct']
        if 'batch_wct' in state:
            self._batch_wct = state['batch_wct']

    @property
    def epoch(self) -> Time[int]:
        """The total epoch count."""
        return self._epoch

    @property
    def batch(self) -> Time[int]:
        """The total batch count."""
        return self._batch

    @property
    def sample(self) -> Time[int]:
        """The total sample count."""
        return self._sample

    @property
    def token(self) -> Time[int]:
        """The total token count."""
        return self._token

    @property
    def batch_in_epoch(self) -> Time[int]:
        """The batch count in the current epoch (resets at 0 at the beginning of every epoch)."""
        return self._batch_in_epoch

    @property
    def sample_in_epoch(self) -> Time[int]:
        """The sample count in the current epoch (resets at 0 at the beginning of every epoch)."""
        return self._sample_in_epoch

    @property
    def token_in_epoch(self) -> Time[int]:
        """The token count in the current epoch (resets at 0 at the beginning of every epoch)."""
        return self._token_in_epoch

    @property
    def total_wct(self) -> datetime.timedelta:
        """The wall-clock duration (in seconds) from the beginning of training."""
        return self._total_wct

    @property
    def epoch_wct(self) -> datetime.timedelta:
        """The wall-clock duration (in seconds) for the current epoch."""
        return self._epoch_wct

    @property
    def batch_wct(self) -> datetime.timedelta:
        """The wall-clock duration (in seconds) for the last batch."""
        return self._batch_wct

    def get(self, unit: Union[str, TimeUnit]) -> Time[int]:
        """Returns the current time in the specified unit.

        Args:
            unit (str | TimeUnit): The desired unit.

        Returns:
            Time: The current time, in the specified unit.
        """
        unit = TimeUnit(unit)
        if unit == TimeUnit.EPOCH:
            return self.epoch
        if unit == TimeUnit.BATCH:
            return self.batch
        if unit == TimeUnit.SAMPLE:
            return self.sample
        if unit == TimeUnit.TOKEN:
            return self.token
        raise ValueError(f'Invalid unit: {unit}')

    def _parse(self, other: object) -> Time:
        # parse ``other`` into a Time object
        if isinstance(other, Time):
            return other
        if isinstance(other, str):
            other_parsed = Time.from_timestring(other)
            return other_parsed

        raise TypeError(f'Cannot convert type {other} to {self.__class__.__name__}')

    def __eq__(self, other: object):
        if not isinstance(other, (Time, Timestamp, str)):
            return NotImplemented
        if isinstance(other, Timestamp):
            return self.state_dict() == other.state_dict()
        other = self._parse(other)
        self_counter = self.get(other.unit)
        return self_counter == other

    def __ne__(self, other: object):
        if not isinstance(other, (Time, Timestamp, str)):
            return NotImplemented
        if isinstance(other, Timestamp):
            return self.state_dict() != other.state_dict()
        other = self._parse(other)
        self_counter = self.get(other.unit)
        return self_counter != other

    def __lt__(self, other: object):
        if not isinstance(other, (Time, str)):
            return NotImplemented
        other = self._parse(other)
        self_counter = self.get(other.unit)
        return self_counter < other

    def __le__(self, other: object):
        if not isinstance(other, (Time, str)):
            return NotImplemented
        other = self._parse(other)
        self_counter = self.get(other.unit)
        return self_counter <= other

    def __gt__(self, other: object):
        if not isinstance(other, (Time, str)):
            return NotImplemented
        other = self._parse(other)
        self_counter = self.get(other.unit)
        return self_counter > other

    def __ge__(self, other: object):
        if not isinstance(other, (Time, str)):
            return NotImplemented
        other = self._parse(other)
        self_counter = self.get(other.unit)
        return self_counter >= other

    def to_next_batch(
        self,
        samples: Union[int, Time] = 0,
        tokens: Union[int, Time] = 0,
        duration: Optional[datetime.timedelta] = None,
    ):
        """Create a new :class:`.Timestamp`, advanced to the next batch.

        Equivalent to:

        .. testsetup::

            from composer.core.time import Timestamp
            import datetime

            timestamp = Timestamp()
            samples = 1
            tokens = 2
            duration = datetime.timedelta(seconds=0)

        .. doctest::

            >>> timestamp.copy(
            ...     batch=timestamp.batch + 1,
            ...     batch_in_epoch=timestamp.batch_in_epoch + 1,
            ...     sample=timestamp.sample + samples,
            ...     sample_in_epoch=timestamp.sample_in_epoch + samples,
            ...     token = timestamp.token + tokens,
            ...     token_in_epoch=timestamp.token_in_epoch + tokens,
            ...     total_wct=timestamp.total_wct + duration,
            ...     epoch_wct=timestamp.epoch_wct + duration,
            ...     batch_wct=duration,
            ... )
            Timestamp(...)

        .. note::

            For accurate time tracking, when doing distributed training, the ``samples`` and ``tokens`` should be
            the total across all ranks for the given batch. This method will not accumulate these counts automatically.
            If per-rank sample and token counts are provided, these counts will differ across ranks, which could lead
            towards inconsistent behavior by :class:`.Algorithm` or :class:`.Callback` instances that use these counts.

        Args:
            samples (int | Time, optional): The number of samples trained in the batch. Defaults to 0.
            tokens (int | Time, optional): The number of tokens trained in the batch. Defaults to 0.
            duration (datetime.timedelta, optional): The duration to train the batch.
        """
        if duration is None:
            duration = datetime.timedelta(seconds=0)
        return self.copy(
            batch=self.batch + 1,
            batch_in_epoch=self.batch_in_epoch + 1,
            sample=self.sample + samples,
            sample_in_epoch=self.sample_in_epoch + samples,
            token=self.token + tokens,
            token_in_epoch=self.token_in_epoch + tokens,
            total_wct=self.total_wct + duration,
            epoch_wct=self.epoch_wct + duration,
            batch_wct=duration,
        )

    def to_next_epoch(self):
        """Create a new :class:`.Timestamp`, advanced to the next epoch.

        Equivalent to:

        .. testsetup::

            from composer.core.time import Timestamp
            import datetime

            timestamp = Timestamp()

        .. doctest::

            >>> timestamp.copy(
            ...     epoch=timestamp.epoch+1,
            ...     batch_in_epoch=0,
            ...     sample_in_epoch=0,
            ...     token_in_epoch=0,
            ...     epoch_wct=datetime.timedelta(seconds=0),
            ...     batch_wct=datetime.timedelta(seconds=0),
            ... )
            Timestamp(...)

        """
        return self.copy(
            epoch=self.epoch + 1,
            batch_in_epoch=0,
            sample_in_epoch=0,
            token_in_epoch=0,
            epoch_wct=datetime.timedelta(seconds=0),
            batch_wct=datetime.timedelta(seconds=0),
        )

    def copy(
        self,
        epoch: Optional[Union[int, Time[int]]] = None,
        batch: Optional[Union[int, Time[int]]] = None,
        sample: Optional[Union[int, Time[int]]] = None,
        token: Optional[Union[int, Time[int]]] = None,
        batch_in_epoch: Optional[Union[int, Time[int]]] = None,
        sample_in_epoch: Optional[Union[int, Time[int]]] = None,
        token_in_epoch: Optional[Union[int, Time[int]]] = None,
        total_wct: Optional[datetime.timedelta] = None,
        epoch_wct: Optional[datetime.timedelta] = None,
        batch_wct: Optional[datetime.timedelta] = None,
    ) -> Timestamp:
        """Create a copy of the timestamp.

        Any specified values will override the existing values in the returned copy.

        Args:
            epoch (int | Time[int], optional): The epoch.
            batch (int | Time[int], optional): the batch.
            sample (int | Time[int], optional): The sample.
            token (int | Time[int], optional): The token.
            batch_in_epoch (int | Time[int], optional): The batch in the epoch.
            sample_in_epoch (int | Time[int], optional): The sample in the epoch.
            token_in_epoch (int | Time[int], optional): The token in the epoch.
            total_wct (datetime.timedelta, optional): The elapsed duration from the beginning of training.

        Returns:
            Timestamp: A new timestamp instance, created from a copy, but with any specified values
                overriding the existing values.
        """
        return Timestamp(
            epoch=epoch if epoch is not None else self.epoch,
            batch=batch if batch is not None else self.batch,
            sample=sample if sample is not None else self.sample,
            token=token if token is not None else self.token,
            batch_in_epoch=batch_in_epoch if batch_in_epoch is not None else self.batch_in_epoch,
            sample_in_epoch=sample_in_epoch if sample_in_epoch is not None else self.sample_in_epoch,
            token_in_epoch=token_in_epoch if token_in_epoch is not None else self.token_in_epoch,
            total_wct=total_wct if total_wct is not None else self.total_wct,
            epoch_wct=epoch_wct if epoch_wct is not None else self.epoch_wct,
            batch_wct=batch_wct if batch_wct is not None else self.batch_wct,
        )

    def __repr__(self) -> str:
        return (f'Timestamp('
                f'epoch={int(self.epoch)}, '
                f'batch={int(self.batch)}, '
                f'sample={int(self.sample)}, '
                f'token={int(self.token)}, '
                f'batch_in_epoch={int(self.batch_in_epoch)}, '
                f'sample_in_epoch={int(self.sample_in_epoch)}, '
                f'token_in_epoch={int(self.token_in_epoch)}, '
                f'total_wct={repr(self.total_wct)}, '
                f'epoch_wct={repr(self.epoch_wct)}, '
                f'batch_wct={repr(self.batch_wct)}'
                ')')


def ensure_time(maybe_time: Union[Time, str, int], int_unit: Union[TimeUnit, str]) -> Time:
    """Ensure ``maybe_time`` is an instance of :class:`.Time`.

    Args:
        maybe_time (Time | str): A time string, integer, or instance of :class:`.Time`.
        int_unit (TimeUnit | str): The unit to use if ``maybe_time`` is an integer

    Returns:
        Time: An instance of :class:`.Time`.
    """
    if isinstance(maybe_time, str):
        return Time.from_timestring(maybe_time)
    if isinstance(maybe_time, int):
        return Time(maybe_time, int_unit)
    if isinstance(maybe_time, Time):
        return maybe_time
    raise TypeError(f'Unsupported type for ensure_time: {type(maybe_time)}')



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

from typing import List, Sequence, Type
from unittest.mock import Mock

import pytest

from composer import Algorithm, Engine, Event, Logger, State, Trainer
from composer.algorithms import FusedLayerNorm, LowPrecisionLayerNorm, SelectiveBackprop
from composer.algorithms.low_precision_layernorm.low_precision_layernorm import LowPrecisionLayerNorm
from composer.core.passes import sort_to_back, sort_to_front
from composer.devices import DeviceCPU
from tests.common import SimpleModel

from .test_engine import run_event


@pytest.fixture
def always_match_algorithms():
    return [
        Mock(**{
            'match.return.value': True,
            'apply.return_value': n,  # return encodes order
            'interpolate_loss': False,
        }) for n in range(5)
    ]


@pytest.fixture()
def dummy_logger(dummy_state: State):
    return Logger(dummy_state)


def test_register_pass(dummy_state, dummy_logger):

    dummy_algorithm = Mock()
    dummy_algorithm.match.return_value = True
    dummy_algorithm.apply.return_value = 'dummy'

    def insert_dummy_algorithm(algorithms, event):
        algorithms.append(dummy_algorithm)
        return algorithms

    engine = Engine(dummy_state, dummy_logger)
    engine.register_pass(insert_dummy_algorithm)

    trace = engine.run_event(Event.INIT)

    assert 'dummy' in [tr.exit_code for tr in trace.values()]


class TestLIFOPass:

    @pytest.mark.parametrize('event', [
        Event.BEFORE_LOSS,
        Event.BEFORE_BACKWARD,
    ])
    def test_lifo_first_in(self, event: Event, dummy_state: State, dummy_logger: Logger,
                           always_match_algorithms: List[Algorithm]):
        dummy_state.algorithms = always_match_algorithms
        trace = run_event(event, dummy_state, dummy_logger)
        order = [tr.order for tr in trace.values()]
        expected_order = [tr.exit_code for tr in trace.values()]  # use exit_code to uniquely label algos

        assert order == expected_order

    @pytest.mark.parametrize('event', [
        Event.AFTER_LOSS,
        Event.AFTER_BACKWARD,
    ])
    def test_lifo_last_out(self, event: Event, dummy_state: State, always_match_algorithms: List[Algorithm],
                           dummy_logger: Logger):
        dummy_state.algorithms = always_match_algorithms
        trace = run_event(event, dummy_state, dummy_logger)
        order = [tr.order for tr in trace.values()]
        expected_order = list(reversed([tr.exit_code for tr in trace.values()]))

        assert order == expected_order


class TestAlgorithmOrderingPasses:

    @pytest.mark.parametrize('algorithm_cls', [FusedLayerNorm, LowPrecisionLayerNorm])
    def test_algorithm_last(self, algorithm_cls: Type[Algorithm], always_match_algorithms: List[Algorithm],
                            dummy_logger: Logger, dummy_state: State):

        if algorithm_cls == FusedLayerNorm or LowPrecisionLayerNorm:
            pytest.importorskip('apex')

        algorithm = algorithm_cls()
        algorithm.apply = Mock(return_value='algo')
        algorithm.match = Mock(return_value=True)

        algortihms = always_match_algorithms[0:2] + [algorithm] + always_match_algorithms[2:]
        dummy_state._algorithms = algortihms

        trace = run_event(Event.INIT, dummy_state, dummy_logger)

        expected = [0, 1, 2, 3, 4, 'algo']
        actual = [tr.exit_code for tr in trace.values()]

        assert actual == expected

    @pytest.mark.parametrize('algorithm_cls', [SelectiveBackprop])
    def test_algorithm_first(self, algorithm_cls: Type[Algorithm], always_match_algorithms: List[Algorithm],
                             dummy_logger: Logger, dummy_state: State):

        algorithm = algorithm_cls()
        algorithm.apply = Mock(return_value='algo')
        algorithm.match = Mock(return_value=True)

        algorithms = always_match_algorithms[0:2] + [algorithm] + always_match_algorithms[2:]
        dummy_state._algorithms = algorithms

        trace = run_event(Event.INIT, dummy_state, dummy_logger)

        expected = ['algo', 0, 1, 2, 3, 4]
        actual = [tr.exit_code for tr in trace.values()]

        assert actual == expected


class TestSortHelpers:

    def test_sort_to_back(self):
        lst = [1, 'a', 'c', 2, 3.0]
        assert sort_to_back(lst, int) == ['a', 'c', 3.0, 1, 2]

    def test_sort_to_front(self):
        lst = [1, 'a', 'c', 2, 3.0]
        assert sort_to_front(lst, int) == [1, 2, 'a', 'c', 3.0]


def get_default_passes():
    state = State(model=SimpleModel(), device=DeviceCPU(), rank_zero_seed=42, run_name='test_chungoose')
    engine = Engine(state, Logger(state))
    return engine.algorithm_passes


def get_custom_pass():

    def sort_by_name(algorithms: Sequence[Algorithm], event: Event) -> Sequence[Algorithm]:
        return sorted(algorithms, key=lambda x: type(x).__name__)

    return sort_by_name


sort_by_name = get_custom_pass()  # Generate pass object so we can use same ref in tests


class TestTrainerArg:

    @pytest.mark.parametrize(
        'algorithm_passes,expected_passes',
        [
            [None, get_default_passes()],
            [sort_by_name, get_default_passes() + [sort_by_name]],
            [[sort_by_name], get_default_passes() + [sort_by_name]],
            [[sort_by_name, 0], [sort_by_name] + get_default_passes()],  # type: ignore
            [(sort_by_name, 0), [sort_by_name] + get_default_passes()],  # type: ignore
            [[(sort_by_name, 0)], [sort_by_name] + get_default_passes()],  # type: ignore
        ])
    def test_add_pass(self, algorithm_passes, expected_passes):
        trainer = Trainer(model=SimpleModel(), algorithm_passes=algorithm_passes)
        assert trainer.engine.algorithm_passes == expected_passes



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""The ResNet model family is a set of convolutional neural networks described in `Deep Residual Learning for Image
Recognition <https://arxiv.org/abs/1512.03385>`_ (He et al, 2015). ResNets can be used as the base for a variety of
vision tasks. ImageNet ResNets are a subset of the ResNet family which were designed specifically for classification on
the ImageNet dataset.

See the :doc:`Model Card </model_cards/resnet>` for more details.
"""
from composer.models.resnet.model import composer_resnet

__all__ = ['composer_resnet']

_metadata = {
    'resnet18': {
        '_task': 'Image Classification',
        '_dataset': 'ImageNet',
        '_name': 'ResNet18',
        '_quality': 'TBD',
        '_metric': 'Top-1 Accuracy',
        '_ttt': 'TBD',
        '_hparams': 'resnet18.yaml'
    },
    'resnet34': {
        '_task': 'Image Classification',
        '_dataset': 'ImageNet',
        '_name': 'ResNet34',
        '_quality': 'TBD',
        '_metric': 'Top-1 Accuracy',
        '_ttt': 'TBD',
        '_hparams': 'resnet34.yaml'
    },
    'resnet50': {
        '_task': 'Image Classification',
        '_dataset': 'ImageNet',
        '_name': 'ResNet50',
        '_quality': '76.51',
        '_metric': 'Top-1 Accuracy',
        '_ttt': '3h 33m',
        '_hparams': 'resnet50.yaml'
    },
    'resnet101': {
        '_task': 'Image Classification',
        '_dataset': 'ImageNet',
        '_name': 'ResNet101',
        '_quality': '78.10',
        '_metric': 'Top-1 Accuracy',
        '_ttt': '8h 15m',
        '_hparams': 'resnet101.yaml',
    },
    'resnet152': {
        '_task': 'Image Classification',
        '_dataset': 'ImageNet',
        '_name': 'ResNet152',
        '_quality': 'TBD',
        '_metric': 'Top-1 Accuracy',
        '_ttt': 'TBD',
        '_hparams': 'resnet152.yaml'
    }
}



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Core RandAugment code."""

import functools
import textwrap
import weakref
from typing import List, TypeVar

import numpy as np
import torch
import torch.utils.data
from PIL.Image import Image as PillowImage
from torchvision.datasets import VisionDataset

from composer.algorithms.utils import augmentation_sets
from composer.algorithms.utils.augmentation_common import map_pillow_function
from composer.core import Algorithm, Event, State
from composer.datasets.utils import add_vision_dataset_transform
from composer.loggers import Logger

__all__ = ['RandAugment', 'RandAugmentTransform', 'randaugment_image']

ImgT = TypeVar('ImgT', torch.Tensor, PillowImage)


def randaugment_image(img: ImgT,
                      severity: int = 9,
                      depth: int = 2,
                      augmentation_set: List = augmentation_sets['all']) -> ImgT:
    """Randomly applies a sequence of image data augmentations  to an image or batch of images.

    This technique is adapted from `Cubuk et al, 2019 <https://arxiv.org/abs/1909.13719>`_).

    See :class:`.RandAugment` or the :doc:`Method Card </method_cards/randaugment>`
    for details. This function only acts on a single image (or batch of images) per call and
    is unlikely to be used in a training loop. Use :class:`.RandAugmentTransform` to use
    :class:`.RandAugment` as part of a :class:`torchvision.datasets.VisionDataset` ``transform``.

    Example:
        .. testcode::

            import composer.functional as cf

            from composer.algorithms.utils import augmentation_sets

            randaugmented_image = cf.randaugment_image(
                img=image,
                severity=9,
                depth=2,
                augmentation_set=augmentation_sets["all"]
            )

    Args:
        img (PIL.Image.Image | torch.Tensor): Image or batch of images to be RandAugmented.
        severity (int, optional): See :class:`.RandAugment`.
        depth (int, optional): See :class:`.RandAugment`.
        augmentation_set (str, optional): See :class:`.RandAugment`.

    Returns:
        PIL.Image: RandAugmented image.
    """

    def _randaugment_pil_image(img: PillowImage, severity: int, depth: int, augmentation_set: List) -> PillowImage:
        # Iterate over augmentations
        for _ in range(depth):
            aug = np.random.choice(augmentation_set)
            img = aug(img, severity)
        return img

    f_pil = functools.partial(_randaugment_pil_image, severity=severity, depth=depth, augmentation_set=augmentation_set)
    return map_pillow_function(f_pil, img)


class RandAugmentTransform(torch.nn.Module):
    """Wraps :func:`.randaugment_image` in a ``torchvision``-compatible transform.

    See :class:`.RandAugment` or the :doc:`Method Card </method_cards/randaugment>` for more details.

    Example:
        .. testcode::

            import torchvision.transforms as transforms
            from composer.algorithms.randaugment import RandAugmentTransform

            randaugment_transform = RandAugmentTransform(
                severity=9,
                depth=2,
                augmentation_set="all"
            )
            composed = transforms.Compose([
                randaugment_transform,
                transforms.RandomHorizontalFlip()
            ])
            transformed_image = composed(image)

    Args:
        severity (int, optional): See :class:`.RandAugment`.
        depth (int, optional): See :class:`.RandAugment`.
        augmentation_set (str, optional): See
            :class:`.RandAugment`.
    """

    def __init__(self, severity: int = 9, depth: int = 2, augmentation_set: str = 'all'):
        super().__init__()
        if severity < 0 or severity > 10:
            raise ValueError('RandAugment severity value must satisfy 0 ≤ severity ≤ 10')
        if depth < 0:
            raise ValueError('RandAugment depth value must be ≥ 0')
        if augmentation_set not in augmentation_sets.keys():
            raise KeyError(f'RandAugment augmentation_set is not one of {augmentation_sets.keys()}')
        self.severity = severity
        self.depth = depth
        self.augmentation_set = augmentation_sets[augmentation_set]

    def forward(self, img: ImgT) -> ImgT:
        return randaugment_image(img=img,
                                 severity=self.severity,
                                 depth=self.depth,
                                 augmentation_set=self.augmentation_set)


class RandAugment(Algorithm):
    """Randomly applies a sequence of image data augmentations to an image.

    This algorithm (`Cubuk et al, 2019 <https://arxiv.org/abs/1909.13719>`_) runs on
    :attr:`.Event.INIT` to insert a dataset
    transformation. It is a no-op if this algorithm already applied itself on the
    :attr:`.State.train_dataloader.dataset`.

    See the :doc:`Method Card </method_cards/randaugment>` for more details.

    Example:
        .. testcode::

            from composer.algorithms import RandAugment
            from composer.trainer import Trainer

            randaugment_algorithm = RandAugment(
                severity=9,
                depth=2,
                augmentation_set="all"
            )
            trainer = Trainer(
                model=model,
                train_dataloader=train_dataloader,
                eval_dataloader=eval_dataloader,
                max_duration="1ep",
                algorithms=[randaugment_algorithm],
                optimizers=[optimizer]
            )

    Args:
        severity (int, optional): Severity of augmentation operators (between 1 to 10). M
            in the original paper. Default: ``9``.
        depth (int, optional): Depth of augmentation chain. N in the original paper.
            Default: ``2``.
        augmentation_set (str, optional): Must be one of the following options
            as also described in :attr:`.augmentation_primitives.augmentation_sets`:

            * ``"all"``
                Uses all augmentations from the paper.
            * ``"safe"``
                Like ``"all"``, but excludes transforms that are part of
                the ImageNet-C/CIFAR10-C test sets
            * ``"original"``
                Like ``"all"``, but some of the implementations
                are identical to the original Github repository, which contains
                implementation specificities for the augmentations
                ``"color"``, ``"contrast"``, ``"sharpness"``, and ``"brightness"``. The
                original implementations have an intensity sampling scheme that samples a
                value bounded by 0.118 at a minimum, and a maximum value of
                :math:`intensity \\times 0.18 + .1`, which ranges from 0.28 (intensity =
                1) to 1.9 (intensity 10). These augmentations have different effects
                depending on whether they are < 0 or > 0 (or < 1 or > 1).
                ``"all"`` uses implementations of ``"color"``, ``"contrast"``,
                ``"sharpness"``, and ``"brightness"`` that account for diverging effects
                around 0 (or 1).

            Default: ``"all"``.
    """

    def __init__(self, severity: int = 9, depth: int = 2, augmentation_set: str = 'all'):
        if severity < 0 or severity > 10:
            raise ValueError('RandAugment severity value must be 0 ≤ severity ≤ 10')
        if augmentation_set not in augmentation_sets.keys():
            raise KeyError(f'randaugment_augmentation_set is not one of {augmentation_sets.keys()}')
        self.severity = severity
        self.depth = depth
        self.augmentation_set = augmentation_set
        self._transformed_datasets = weakref.WeakSet()

    def match(self, event: Event, state: State) -> bool:
        if event != Event.FIT_START:
            return False
        assert state.dataloader is not None, 'dataloader should be defined on fit start'
        if not isinstance(state.dataloader, torch.utils.data.DataLoader):
            raise TypeError(f'{type(self).__name__} requires a PyTorch dataloader.')
        return state.dataloader.dataset not in self._transformed_datasets

    def apply(self, event: Event, state: State, logger: Logger) -> None:
        ra = RandAugmentTransform(severity=self.severity, depth=self.depth, augmentation_set=self.augmentation_set)
        assert isinstance(state.dataloader, torch.utils.data.DataLoader), 'The dataloader type is checked on match()'
        dataset = state.dataloader.dataset
        if not isinstance(dataset, VisionDataset):
            raise TypeError(
                textwrap.dedent(f"""\
                To use {type(self).__name__}, the dataset must be a
                {VisionDataset.__qualname__}, not {type(dataset).__name__}"""))
        add_vision_dataset_transform(dataset, ra, is_tensor_transform=False)
        self._transformed_datasets.add(dataset)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

from composer import (algorithms, callbacks, core, datasets, devices, functional, loggers, loss, metrics, models, optim,
                      profiler, trainer, utils)


# This very simple test is just to use the above imports, which check and make sure we can import all the top-level
# modules from composer. This is mainly useful for checking that we have correctly conditionally imported all optional
# dependencies.
def test_smoketest():
    assert callbacks
    assert algorithms
    assert core
    assert datasets
    assert devices
    assert functional
    assert loggers
    assert loss
    assert metrics
    assert models
    assert optim
    assert profiler
    assert trainer
    assert utils



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

# Copyright 2022 MosaicML. All Rights Reserved.

from __future__ import annotations

import logging
import warnings
from typing import Callable, Dict, Optional, Sequence, Type, Union

import torch

from composer.models.huggingface import HuggingFaceModel

try:
    from transformers import BertPreTrainedModel
    from transformers.models.bert.modeling_bert import BertIntermediate, BertOutput
    IS_TRANSFORMERS_INSTALLED = True
except ImportError as e:
    IS_TRANSFORMERS_INSTALLED = False

from composer.algorithms.gated_linear_units.gated_linear_unit_layers import BERTGatedFFOutput
from composer.algorithms.warnings import NoEffectWarning
from composer.core import Algorithm, Event, State
from composer.loggers import Logger
from composer.utils import MissingConditionalImportError, module_surgery

log = logging.getLogger(__name__)


def from_BertOutput(layer: torch.nn.Module,
                    module_index: int,
                    act_fn: Callable[[torch.Tensor], torch.Tensor],
                    gated_layer_bias: bool = False,
                    non_gated_layer_bias: bool = False) -> BERTGatedFFOutput:
    """Defines a replacement policy from a :class:`transformers.models.bert.modeling_bert.BertOutput` to a :class:`composer.algorithms.gated_linear_units.gated_linear_unit_layers.BERTGatedFFOutput`"""
    assert isinstance(
        layer, BertOutput
    ), 'The replacement policy requires an instance of transformers.models.bert.modeling_bert.BertOutput for the necessary fields to be defined.'
    return BERTGatedFFOutput(
        d_embed=layer.dense.out_features,  #type: ignore dense.out_features member of BertOutput
        d_ff=layer.dense.in_features,  #type: ignore dense.in_features member of BertOutput
        dropout_rate=layer.dropout.p,  #type: ignore dropout.p member of BertOutput
        act_fn=act_fn,
        layernorm_eps=layer.LayerNorm.eps,  #type: ignore LayerNorm.eps member of BertOutput
        gated_layer_bias=gated_layer_bias,
        non_gated_layer_bias=non_gated_layer_bias)


def from_BertIntermediate(layer: torch.nn.Module, module_index: int) -> torch.nn.Identity:
    """
    Defines a replacement policy from a :class:`transformers.models.bert.modeling_bert.BertIntermediate` to a :class:`torch.nn.Identity`
    The identity effectively acts as no-op.
    """
    return torch.nn.Identity()


def apply_gated_linear_units(model: torch.nn.Module,
                             optimizers: Union[torch.optim.Optimizer, Sequence[torch.optim.Optimizer]],
                             act_fn: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
                             gated_layer_bias: bool = False,
                             non_gated_layer_bias: bool = False) -> None:
    """
    Replaces the Linear layers in the feed-forward network with `Gated Linear Units <https://arxiv.org/abs/2002.05202>`_.

    Args:
        model (`torch.nn.Module`): The model to modify in-place.
        optimizers (`torch.optim.Optimizer` | Sequence[`torch.optim.Optimizer`], optional):
            Existing optimizers bound to ``model.parameters()``. All optimizers that have already been
            constructed with ``model.parameters()`` must be specified here so that
            they will optimize the correct parameters.

            If the optimizer(s) are constructed after calling this function,
            then it is safe to omit this parameter. These optimizers will see the correct
            model parameters.
        act_fn (Callable[torch.Tensor, torch.Tensor], optional): Optionally, the activation function to use. If ``None``, the algorithm will
            use the existing activation function in the model.
        gated_layer_bias (bool, optional): Whether to use biases in the linear layers within the GLU. Default: ``False``.
        non_gated_layer_bias (bool, optional): Whether to use biases in the linear layers within the GLU. Default: ``False``.
    """
    if not IS_TRANSFORMERS_INSTALLED:
        raise MissingConditionalImportError(extra_deps_group='nlp', conda_package='transformers')

    unwrapped_model = model.model if isinstance(model, HuggingFaceModel) else model

    # ensure that the model is an instance of a Hugging Face BertPreTrainedModel class, since our replacement policy is only defined for BERTs
    if not isinstance(unwrapped_model, BertPreTrainedModel):
        raise TypeError(
            'Gated Linear Units only has a surgery policy defined for subclasses of transformers.BertPreTrainedModel')

    if act_fn is None:
        intermediate_modules = {module for module in model.modules() if isinstance(module, BertIntermediate)}
        if len(intermediate_modules) == 0:
            warnings.warn(
                NoEffectWarning('No instances of BertIntermediate were found so Gated Linear Units will be skipped '
                                'as no modules can be replaced. This is likely because Gated Linear Units has already '
                                'been applied to this model.'))

        # get the activation functions used
        act_fns = {module.intermediate_act_fn for module in intermediate_modules}
        num_act_fns = len({type(act_fn) for act_fn in act_fns})
        if num_act_fns == 0:
            raise ValueError('Tried to get the activation function from the model, but none were found. '
                             'Please specify `act_fn` manually to use Gated Linear Units.')
        elif num_act_fns > 1:
            raise ValueError('Tried to get the activation function from the model, but multiple different '
                             'functions are used. This is currently unsupported with Gated Linear Units. '
                             'Please either use one activation function in BertIntermediate modules or '
                             'specify `act_fn` to manually override activation functions.')

        # since our set is of 1, let's extract the activation function
        act_fn = next(iter(act_fns))  # type: ignore will fail below if None

        if act_fn is None:
            raise ValueError(
                'Found activation function was None. If this is an error, please manually specify `act_fn`.')

    # now that we know the act fn, bind a few parameters of the replacement function
    def from_bound_BertOutput(layer: torch.nn.Module, module_index: int) -> BERTGatedFFOutput:
        return from_BertOutput(layer=layer,
                               module_index=module_index,
                               act_fn=act_fn,
                               gated_layer_bias=gated_layer_bias,
                               non_gated_layer_bias=non_gated_layer_bias)

    # prepare the replacement policy and perform replacement
    policy: Dict[Type[torch.nn.Module], module_surgery.ReplacementFunction] = {
        BertIntermediate: from_BertIntermediate,
        BertOutput: from_bound_BertOutput
    }
    replaced_instances = module_surgery.replace_module_classes(module=model, optimizers=optimizers, policies=policy)
    if len(replaced_instances) == 0:
        warnings.warn(
            NoEffectWarning('No instances of BertIntermediate and BertOutput were found so no modules were replaced.'))
    log.info(
        f'Successfully replaced {len(replaced_instances)} of BertIntermediate and BertOutput with a GatedLinearUnit.')


class GatedLinearUnits(Algorithm):
    """Replaces all instances of Linear layers in the feed-forward subnetwork with a `Gated Linear Unit <https://arxiv.org/abs/2002.05202>`_.
    The Gated Linear Units provide a more expressive form for the same number of parameters, and a slight degradation to throughput.

    Runs on :attr:`.Event.INIT`, so it can swap the Linear layers in the FFN for GLUs before the model is DDP wrapped.

    Args:
        act_fn (Callable[[torch.Tensor], torch.Tensor], optional): Optionally, the activation function to use. If ``None``, the algorithm will
            use the existing activation function in the model.
        gated_layer_bias (bool, optional): Whether to use biases in the linear layers within the GLU. Default: ``False``.
        non_gated_layer_bias (bool, optional): Whether to use biases in the linear layers within the GLU. Default: ``False``.

    Example:
        .. testsetup::

           from tests.common.models import configure_tiny_bert_hf_model
           from tests.common.datasets import dummy_bert_lm_dataloader

           model, train_dataloader = configure_tiny_bert_hf_model(), dummy_bert_lm_dataloader()

        .. testcode::

           from composer.algorithms import GatedLinearUnits

           algorithm = GatedLinearUnits()
           trainer = Trainer(
               model=model,
               train_dataloader=train_dataloader,
               max_duration="1ep",
               algorithms=[algorithm],
               optimizers=[optimizer]
           )
    """

    def __init__(self,
                 act_fn: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
                 gated_layer_bias: bool = False,
                 non_gated_layer_bias: bool = False):
        if not IS_TRANSFORMERS_INSTALLED:
            raise MissingConditionalImportError(extra_deps_group='nlp', conda_package='transformers')
        self.act_fn = act_fn
        self.gated_layer_bias = gated_layer_bias
        self.non_gated_layer_bias = non_gated_layer_bias

    def __repr__(self) -> str:
        act_fn = 'act_fn' if self.act_fn else None
        return f'{self.__class__.__name__}(act_fn={act_fn},gated_layer_bias={self.gated_layer_bias},non_gated_layer_bias={self.non_gated_layer_bias})'

    @staticmethod
    def required_on_load() -> bool:
        return True

    def match(self, event: Event, state: State) -> bool:
        del state  # unused
        return event == Event.INIT

    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        del event, logger  # unused
        apply_gated_linear_units(model=state.model,
                                 optimizers=state.optimizers,
                                 act_fn=self.act_fn,
                                 gated_layer_bias=self.gated_layer_bias,
                                 non_gated_layer_bias=self.non_gated_layer_bias)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""A simple convolutional neural network extending :class:`.ComposerClassifier`."""

from typing import List, Optional, Sequence, Union

import torch
import torch.nn as nn
from torch.nn import functional as F

from composer.models.initializers import Initializer
from composer.models.tasks import ComposerClassifier

__all__ = ['Model', 'mnist_model']


class Model(nn.Module):
    """Toy convolutional neural network architecture in pytorch for MNIST."""

    def __init__(self, initializers: Sequence[Union[str, Initializer]], num_classes: int = 10):
        super().__init__()

        self.num_classes = num_classes

        for initializer in initializers:
            initializer = Initializer(initializer)
            self.apply(initializer.get_initializer())

        self.conv1 = nn.Conv2d(1, 16, (3, 3), padding=0)
        self.conv2 = nn.Conv2d(16, 32, (3, 3), padding=0)
        self.bn = nn.BatchNorm2d(32)
        self.fc1 = nn.Linear(32 * 16, 32)
        self.fc2 = nn.Linear(32, num_classes)

    def forward(self, x):
        out = self.conv1(x)
        out = F.relu(out)
        out = self.conv2(out)
        out = self.bn(out)
        out = F.relu(out)
        out = F.adaptive_avg_pool2d(out, (4, 4))
        out = torch.flatten(out, 1, -1)
        out = self.fc1(out)
        out = F.relu(out)
        return self.fc2(out)


def mnist_model(num_classes: int = 10, initializers: Optional[List[Initializer]] = None):
    """Helper function to create a :class:`.ComposerClassifier` with a simple convolutional neural network.

    Args:
        num_classes (int, optional): The number of classes. Needed for classification tasks. Default: ``10``
        initializers (List[Initializer], optional): list of Initializers
            for the model. ``None`` for no initialization. Default: ``None``

    Returns:
        ComposerModel: instance of :class:`.ComposerClassifier` with a simple MNIST model.

    Example:

    .. testcode::

        from composer.models import mnist_model

        model = mnist_model()
    """

    if initializers is None:
        initializers = []

    model = Model(initializers, num_classes)
    composer_model = ComposerClassifier(module=model)
    return composer_model



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Progressively freeze the layers of the network during training, starting with the earlier layers.

See the :doc:`Method Card </method_cards/layer_freezing>` for more details.
"""

from composer.algorithms.layer_freezing.layer_freezing import LayerFreezing as LayerFreezing
from composer.algorithms.layer_freezing.layer_freezing import freeze_layers as freeze_layers

__all__ = ['LayerFreezing', 'freeze_layers']



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""A wrapper around `timm.create_model() <https://rwightman.github.io/pytorch-image-models/#load-a-pretrained-model>`_
used to create :class:`.ComposerClassifier`."""

from composer.models.timm.model import composer_timm as composer_timm

__all__ = ['composer_timm']



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Clips all gradients in a model based on their values, their norms,
and their parameters' norms.

See the :doc:`Method Card </method_cards/gradient_clipping>` for more details.
"""

from composer.algorithms.gradient_clipping.gradient_clipping import GradientClipping, apply_gradient_clipping

__all__ = ['GradientClipping', 'apply_gradient_clipping']



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Composer Profiler."""

from __future__ import annotations

import logging
import pathlib
from typing import TYPE_CHECKING, Callable, Dict, List, Sequence, Tuple, Union

from composer.profiler.marker import Marker
from composer.profiler.profiler_action import ProfilerAction
from composer.profiler.system_profiler import SystemProfiler
from composer.profiler.torch_profiler import TorchProfiler
from composer.profiler.trace_handler import TraceHandler
from composer.utils import ensure_tuple

if TYPE_CHECKING:
    from composer.core import Callback, State

__all__ = ['Profiler']

log = logging.getLogger(__name__)


class Profiler:
    """Composer Profiler.

    See the :doc:`Profiling Guide </trainer/performance_tutorials/profiling>` for additional information.

    Args:
        schedule ((State) -> ProfilerAction): The profiling scheduling function.

            It takes the training state and returns a :class:`.ProfilerAction`.
            For convenience, Composer includes a :meth:`~composer.profiler.cyclic_schedule.cyclic_schedule` helper.

            .. testsetup::

                from composer.profiler import Profiler, cyclic_schedule

                original_profiler_init = Profiler.__init__

                def new_profiler_init(self, dummy_ellipsis=None, **kwargs):
                    if 'trace_handlers' not in kwargs:
                        kwargs['trace_handlers'] = []
                    original_profiler_init(self, **kwargs)

                Profiler.__init__ = new_profiler_init

            .. testcode::

                from composer.profiler import Profiler, cyclic_schedule

                profiler = Profiler(
                    ...,
                    schedule=cyclic_schedule(
                        skip_first=1,
                        wait=0,
                        warmup=1,
                        active=4,
                        repeat=1,
                    ),
                )

        trace_handlers (TraceHandler | Sequence[TraceHandler]): Trace handlers which record and
            save profiling data to traces.
        sys_prof_cpu (bool, optional): Whether to record cpu statistics. (default: ``True``).
        sys_prof_memory (bool, optional): Whether to record memory statistics. (default: ``False``).
        sys_prof_disk (bool, optional): Whether to record disk statistics. (default: ``False``).
        sys_prof_net (bool, optional): Whether to record network statistics. (default: ``False``).
        sys_prof_stats_thread_interval_seconds (float, optional): Interval to record stats, in seconds.
            (default: ``0.5``).
        torch_prof_folder (str, optional): See :class:`~composer.profiler.torch_profiler.TorchProfiler`.
        torch_prof_filename (str, optional): See :class:`~composer.profiler.torch_profiler.TorchProfiler`.
        torch_prof_remote_file_name (str, optional): See :class:`~composer.profiler.torch_profiler.TorchProfiler`.
        torch_prof_overwrite (bool, optional): See :class:`~composer.profiler.torch_profiler.TorchProfiler`.
        torch_prof_use_gzip (bool, optional): See :class:`~composer.profiler.torch_profiler.TorchProfiler`.
        torch_prof_record_shapes (bool, optional): See :class:`~composer.profiler.torch_profiler.TorchProfiler`.
        torch_prof_profile_memory (bool, optional): See :class:`~composer.profiler.torch_profiler.TorchProfiler`.
        torch_prof_with_stack (bool, optional): See :class:`~composer.profiler.torch_profiler.TorchProfiler`.
        torch_prof_with_flops (bool, optional): See :class:`~composer.profiler.torch_profiler.TorchProfiler`.
        torch_prof_num_traces_to_keep (int, optional): See :class:`~composer.profiler.torch_profiler.TorchProfiler`.
    """

    def __init__(
        self,
        schedule: Callable[[State], ProfilerAction],
        trace_handlers: List[TraceHandler],
        sys_prof_cpu: bool = True,
        sys_prof_memory: bool = False,
        sys_prof_disk: bool = False,
        sys_prof_net: bool = False,
        sys_prof_stats_thread_interval_seconds: float = 0.5,
        torch_prof_folder: str = '{run_name}/torch_traces',
        torch_prof_filename: str = 'rank{rank}.{batch}.pt.trace.json',
        torch_prof_remote_file_name: str = '{run_name}/torch_traces/rank{rank}.{batch}.pt.trace.json',
        torch_prof_overwrite: bool = False,
        torch_prof_use_gzip: bool = False,
        torch_prof_record_shapes: bool = False,
        torch_prof_profile_memory: bool = True,
        torch_prof_with_stack: bool = False,
        torch_prof_with_flops: bool = True,
        torch_prof_num_traces_to_keep: int = -1,
    ) -> None:
        self._names_to_markers: Dict[str, Marker] = {}
        self._trace_handlers = list(ensure_tuple(trace_handlers))
        self.schedule = schedule
        self.state = None
        self._callbacks: List[Callback] = []

        if sys_prof_cpu or sys_prof_memory or sys_prof_disk or sys_prof_net:
            self._callbacks.append(
                SystemProfiler(profile_cpu=sys_prof_cpu,
                               profile_memory=sys_prof_memory,
                               profile_disk=sys_prof_disk,
                               profile_net=sys_prof_net,
                               stats_thread_interval_seconds=sys_prof_stats_thread_interval_seconds))

        if torch_prof_record_shapes or torch_prof_profile_memory or torch_prof_with_stack or torch_prof_with_flops:
            self._callbacks.append(
                TorchProfiler(filename=torch_prof_filename,
                              folder=torch_prof_folder,
                              remote_file_name=torch_prof_remote_file_name,
                              num_traces_to_keep=torch_prof_num_traces_to_keep,
                              overwrite=torch_prof_overwrite,
                              record_shapes=torch_prof_record_shapes,
                              profile_memory=torch_prof_profile_memory,
                              use_gzip=torch_prof_use_gzip,
                              with_stack=torch_prof_with_stack,
                              with_flops=torch_prof_with_flops))

    def bind_to_state(
        self,
        state: State,
    ):
        """Bind the profiler to the ``state``.

        .. note::

            The :class:`.Trainer` automatically invokes this method.

        Args:
            state (State): The training state.
        """
        self.state = state
        self.state.callbacks.extend(self._callbacks)
        self.state.callbacks.extend(self._trace_handlers)

    @property
    def trace_handlers(self):
        """Profiler trace handlers."""
        return self._trace_handlers

    @trace_handlers.setter
    def trace_handlers(self, trace_handlers: Union[TraceHandler, Sequence[TraceHandler]]):
        """Profiler trace handlers."""
        self._trace_handlers[:] = ensure_tuple(trace_handlers)

    def record_chrome_json_trace_file(self, filepath: Union[str, pathlib.Path]):
        """Record trace events in Chrome JSON format in the trace handlers.

        See `this document <https://docs.google.com/document/d/1CvAClvFfyA5R-PhYUmn5OOQtYMH4h6I0nSsKchNAySU/preview>`_
        for more information about Chrome JSON format.

        .. note::

            For custom profiling, it is recommended to use :meth:`marker` instead of manually creating a Chrome JSON
            trace file. By default, the Composer Profiler will automatically saving :class:`.Marker` events in Chrome
            JSON format.

            This method exists for external profilers that natively record events in Chrome JSON format (such as the
            :class:`~composer.profiler.torch_profiler.TorchProfiler`). These profilers can use this method to route
            their profiling traces to the Composer profiler :attr:`~trace_handlers` so events from both the Composer
            Profiler and external profilers are recorded in the same trace file.
        """
        for recorder in self.trace_handlers:
            recorder.process_chrome_json_trace_file(pathlib.Path(filepath))

    def marker(
            self,
            name: str,
            actions: Sequence[ProfilerAction] = (ProfilerAction.WARMUP, ProfilerAction.ACTIVE,
                                                 ProfilerAction.ACTIVE_AND_SAVE),
            record_instant_on_start: bool = False,
            record_instant_on_finish: bool = False,
            categories: Union[List[str], Tuple[str, ...]] = (),
    ) -> Marker:
        """Create and get an instance of a :class:`.Marker`.

        If a :class:`.Marker` with the specified ``name`` does not already exist, it will be created.
        Otherwise, the existing instance will be returned.

        .. note::

            :meth:`.Profiler.marker()` should be used to construct markers.  :class:`.Marker` **should not** be
            instantiated directly by the user.

            For example:

            .. testsetup:: composer.profiler.profiler.Profiler.marker

                from composer.profiler import Profiler, cyclic_schedule

                profiler = Profiler(schedule=cyclic_schedule(), trace_handlers=[])
                profiler.bind_to_state(state)
                state.profiler = profiler

            .. doctest:: composer.profiler.profiler.Profiler.marker

                >>> marker = profiler.marker("foo")
                >>> marker
                <composer.profiler.marker.Marker object at ...>

        Please see :meth:`.Marker.start()` and :meth:`.Marker.finish()` for usage on creating markers to measure duration events,
        :meth:`.Marker.instant()` for usage on creating markers to mark instant events and :meth:`.Marker.counter()` for usage on
        creating markers for counting.

        Args:
            name (str): The name for the :class:`.Marker`.
            actions (Sequence[ProfilerAction], optional): :class:`.ProfilerAction` states to record on.
                Defaults to (:attr:`~.ProfilerAction.WARMUP`, :attr:`~.ProfilerAction.ACTIVE`,
                :attr:`~.ProfilerAction.ACTIVE_AND_SAVE`).
            record_instant_on_start (bool, optional): Whether to record an instant event whenever the marker is started.
                Defaults to ``False``.
            record_instant_on_finish (bool, optional): Whether to record an instant event whenever the marker is finished.
                Defaults to ``False``.
            categories (Union[List[str], Tuple[str, ...]], optional): Categories for this marker. Defaults to ``None``.

        Returns:
            Marker: Marker instance.
        """
        if self.state is None:
            raise RuntimeError('Profiler.bind_to_state() must be invoked before the Profiler can be used.')
        if name not in self._names_to_markers:

            def should_record(state: State) -> bool:
                return self.schedule(state) in actions

            self._names_to_markers[name] = Marker(
                state=self.state,
                trace_handlers=self.trace_handlers,
                name=name,
                should_record=should_record,
                record_instant_on_start=record_instant_on_start,
                record_instant_on_finish=record_instant_on_finish,
                categories=categories,
            )
        self._names_to_markers[name].categories = categories
        return self._names_to_markers[name]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Module Initializers."""

from typing import Callable

import torch
from torch import nn as nn

from composer.utils import StringEnum


class Initializer(StringEnum):
    """Sets the initialization scheme for different layers of a PyTorch model."""
    KAIMING_NORMAL = 'kaiming_normal'
    KAIMING_UNIFORM = 'kaiming_uniform'
    BN_UNIFORM = 'bn_uniform'
    BN_ONES = 'bn_ones'
    XAVIER_UNIFORM = 'xavier_uniform'
    XAVIER_NORMAL = 'xavier_normal'
    LINEAR_LOG_CONSTANT_BIAS = 'linear_log_constant_bias'

    def get_initializer(self) -> Callable[[torch.nn.Module], None]:
        """Get the initializer function.

        Returns:
            (torch.nn.Module) -> None: The initializer function.
        """

        def kaiming_normal(w: nn.Module):
            if isinstance(w, torch.nn.Linear) or isinstance(w, torch.nn.Conv2d):
                torch.nn.init.kaiming_normal_(w.weight)

        def kaiming_uniform(w: nn.Module):
            if isinstance(w, torch.nn.Linear) or isinstance(w, torch.nn.Conv2d):
                torch.nn.init.kaiming_uniform_(w.weight)

        def xavier_uniform(w: nn.Module):
            if isinstance(w, torch.nn.Linear) or isinstance(w, torch.nn.Conv2d):
                torch.nn.init.xavier_uniform_(w.weight)

        def xavier_normal(w: nn.Module):
            if isinstance(w, torch.nn.Linear) or isinstance(w, torch.nn.Conv2d):
                torch.nn.init.xavier_normal_(w.weight)

        def bn_ones(w: nn.Module):
            if isinstance(w, torch.nn.BatchNorm2d):
                w.weight.data = torch.ones_like(w.weight.data)
                w.bias.data = torch.zeros_like(w.bias.data)

        def bn_uniform(w: nn.Module):
            if isinstance(w, torch.nn.BatchNorm2d):
                w.weight.data = torch.rand(w.weight.data.shape)
                w.bias.data = torch.zeros_like(w.bias.data)

        def linear_log_constant_bias(w: nn.Module):
            if isinstance(w, torch.nn.Linear):
                w.bias.data = torch.ones(w.bias.shape) * -torch.log(torch.tensor(w.bias.shape[0]))

        initializer_dict = {
            'kaiming_normal': kaiming_normal,
            'kaiming_uniform': kaiming_uniform,
            'bn_uniform': bn_uniform,
            'bn_ones': bn_ones,
            'xavier_uniform': xavier_uniform,
            'xavier_normal': xavier_normal,
            'linear_log_constant_bias': linear_log_constant_bias
        }
        if self.value not in initializer_dict:
            raise ValueError(f"Initializer '{self.value}' not found.")
        return initializer_dict[self.value]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Train models with flexible insertion of algorithms."""

from composer.trainer.trainer import Trainer

__all__ = ['Trainer']



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Replaces all instances of `torch.nn.LayerNorm` with a `apex.normalization.fused_layer_norm.FusedLayerNorm
<https://nvidia.github.io/apex/layernorm.html>`_.

By fusing multiple kernel launches into one, this usually improves GPU utilization.

See the :doc:`Method Card </method_cards/fused_layernorm>` for more details.
"""

from composer.algorithms.fused_layernorm.fused_layernorm import FusedLayerNorm, apply_fused_layernorm

__all__ = ['FusedLayerNorm', 'apply_fused_layernorm']
